from dataclasses import dataclass

from sqlalchemy import func, select
from sqlalchemy.sql.selectable import Select


@dataclass
class PageCalc:
    limit: int
    offset: int
    next_page: int


@dataclass
class Pagination:
    total: int
    limit: int
    page: int


async def get_total_async(session, Model):
    """Should be made in a context manager"""
    stmt = select(func.count(Model.id))
    _total = await session.execute(stmt)
    total = _total.scalar()

    return total


def get_total(session, Model):
    """Should be made in a context manager"""
    stmt = select(func.count(Model.id))
    _total = session.execute(stmt)
    total = _total.scalar()

    return total


def pagination(s: Select, p: Pagination):

    offset = p.limit * (p.page - 1)
    next_page = p.page + 1
    next_offset = p.limit * p.page
    if next_offset >= p.total:
        next_page = -1
    stmt = s.limit(p.limit).offset(offset)
    return stmt, next_page


def calculate_page(total, limit, page) -> PageCalc:
    offset = limit * (page - 1)
    next_page = page + 1
    next_offset = limit * page
    if next_offset >= total:
        next_page = -1
    return PageCalc(limit=limit, offset=offset, next_page=next_page)



from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import List, Optional

from labfunctions import defaults, errors
from labfunctions.executors.execid import ExecID
from labfunctions.hashes import generate_random
from labfunctions.types import ExecutionNBTask, ExecutionResult, NBTask, ServerSettings
from labfunctions.types.runtimes import RuntimeData
from labfunctions.utils import get_version, today_string

WFID_PREFIX = "tmp"


def execid_for_build(size=defaults.EXECID_LEN) -> str:
    return ExecID().firm_with(ExecID.types.build)


def dummy_wfid():
    return f"{WFID_PREFIX}{generate_random(defaults.WFID_LEN - len(WFID_PREFIX))}"


def create_dummy_ctx(projectid, execid=None) -> ExecutionNBTask:
    dummy_id = execid or f"{WFID_PREFIX}.{generate_random(defaults.WFID_LEN)}"
    # pd = ProjectData(name=pname, projectid=projectid)
    task = NBTask(nb_name="welcome", params={})

    ctx = create_notebook_ctx(projectid, task, execid=dummy_id)
    return ctx


def prepare_runtime(runtime: Optional[RuntimeData] = None, gpu_support=False) -> str:
    if not runtime:
        version = get_version()
        _runtime = f"{defaults.DOCKERFILE_IMAGE}:{version}"
        if gpu_support:
            _runtime = f"{defaults.DOCKERFILE_IMAGE_GPU}:{version}"
    else:
        _runtime = f"{runtime.docker_name}:{runtime.version}"
        if runtime.registry:
            _runtime = f"{runtime.registry}/{runtime}"
    return _runtime


def create_notebook_ctx(
    projectid: str,
    task: NBTask,
    execid: Optional[str] = None,
    runtime: Optional[RuntimeData] = None,
    wfid: Optional[str] = None,
) -> ExecutionNBTask:
    """It creates the execution context of a notebook based on project and workflow data"""
    # root = Path.cwd()
    root = Path(defaults.NOTEBOOKS_DIR)
    today = today_string(format_="day")
    _now = datetime.utcnow().isoformat()

    wfid = wfid or dummy_wfid()
    execid = execid or str(ExecID())

    _params = deepcopy(task.params)
    _params["WFID"] = wfid
    _params["EXECID"] = execid
    _params["NOW"] = _now

    nb_filename = f"{task.nb_name}.ipynb"

    papermill_input = str(root / nb_filename)

    output_dir = f"{defaults.NB_OUTPUTS}/ok/{today}"
    error_dir = f"{defaults.NB_OUTPUTS}/errors/{today}"

    output_name = f"{wfid}.{task.nb_name}.{execid}.ipynb"
    _runtime = prepare_runtime(runtime, task.gpu_support)
    machine = task.machine or defaults.MACHINE_TYPE
    cluster = task.cluster or defaults.CLUSTER_NAME

    return ExecutionNBTask(
        projectid=projectid,
        wfid=wfid,
        execid=execid,
        nb_name=task.nb_name,
        runtime=_runtime,
        machine=task.machine,
        cluster=cluster,
        params=_params,
        pm_input=str(papermill_input),
        pm_output=f"{output_dir}/{output_name}",
        output_name=output_name,
        output_dir=output_dir,
        error_dir=error_dir,
        today=today,
        timeout=task.timeout,
        gpu_support=task.gpu_support,
        created_at=_now,
        notifications_ok=task.notifications_ok,
        notifications_fail=task.notifications_fail,
    )


def make_error_result(ctx: ExecutionNBTask, elapsed) -> ExecutionResult:
    result = ExecutionResult(
        wfid=ctx.wfid,
        execid=ctx.execid,
        projectid=ctx.projectid,
        name=ctx.nb_name,
        params=ctx.params,
        input_=ctx.pm_input,
        output_dir=ctx.output_dir,
        output_name=ctx.output_name,
        error_dir=ctx.error_dir,
        error=True,
        elapsed_secs=round(elapsed, 2),
        created_at=ctx.created_at,
    )
    return result



import json
import logging
import os
import shutil
import time
from pathlib import Path
from typing import Union

from labfunctions import client, defaults
from labfunctions.conf import load_client
from labfunctions.types import ExecutionNBTask, ExecutionResult, NBTask

from .nbtask_base import NBTaskLocal

# from labfunctions.notebooks import nb_job_executor


def local_exec_env() -> ExecutionResult:
    """
    Control the notebook execution.
    TODO: implement notifications
    TODO: base executor class?
    """
    # Init
    nbclient = client.from_env()
    runner = NBTaskLocal(nbclient)
    ctx_str = os.getenv(defaults.EXECUTIONTASK_VAR)

    etask = ExecutionNBTask(**json.loads(ctx_str))
    result = runner.run(etask)

    if not os.getenv("DEBUG"):
        runner.register(result)

    return result



"""init

Revision ID: 0000
Revises: 
Create Date: 2022-04-05 17:52:16.030969

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0000"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "nb_auth_user",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("username", sa.String(), nullable=False),
        sa.Column("password", sa.BINARY(), nullable=True),
        sa.Column("is_superuser", sa.Boolean(), nullable=False),
        sa.Column("is_active", sa.Boolean(), nullable=False),
        sa.Column("scopes", sa.String(), nullable=False),
        sa.Column(
            "created_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.Column(
            "updated_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_nb_auth_user_username"), "nb_auth_user", ["username"], unique=True
    )
    op.create_table(
        "nb_project",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("projectid", sa.String(length=16), nullable=False),
        sa.Column("name", sa.String(length=128), nullable=False),
        sa.Column("private_key", sa.BINARY(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("repository", sa.String(length=2048), nullable=True),
        sa.Column("owner_id", sa.BigInteger(), nullable=False),
        sa.Column("agent_id", sa.BigInteger(), nullable=True),
        sa.Column(
            "created_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.Column(
            "updated_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.ForeignKeyConstraint(["agent_id"], ["nb_auth_user.id"], ondelete="SET NULL"),
        sa.ForeignKeyConstraint(["owner_id"], ["nb_auth_user.id"], ondelete="SET NULL"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_index(
        op.f("ix_nb_project_projectid"), "nb_project", ["projectid"], unique=True
    )
    op.create_table(
        "nb_history",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("wfid", sa.String(length=24), nullable=True),
        sa.Column("execid", sa.String(length=24), nullable=True),
        sa.Column("nb_name", sa.String(), nullable=False),
        sa.Column("result", sa.JSON(), nullable=False),
        sa.Column("elapsed_secs", sa.Float(), nullable=False),
        sa.Column("status", sa.Integer(), nullable=True),
        sa.Column(
            "created_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.Column("project_id", sa.String(length=16), nullable=True),
        sa.ForeignKeyConstraint(
            ["project_id"], ["nb_project.projectid"], ondelete="SET NULL"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_nb_history_created_at"), "nb_history", ["created_at"], unique=False
    )
    op.create_index(
        op.f("ix_nb_history_status"), "nb_history", ["status"], unique=False
    )
    op.create_table(
        "nb_projects_users",
        sa.Column("project_id", sa.String(length=16), nullable=True),
        sa.Column("user_id", sa.BigInteger(), nullable=True),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["nb_project.projectid"],
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["nb_auth_user.id"],
        ),
    )
    op.create_table(
        "nb_runtime",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("docker_name", sa.String(), nullable=False),
        sa.Column("version", sa.String(), nullable=False),
        sa.Column(
            "created_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.Column("project_id", sa.String(length=16), nullable=True),
        sa.ForeignKeyConstraint(
            ["project_id"], ["nb_project.projectid"], ondelete="SET NULL"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_nb_runtime_created_at"), "nb_runtime", ["created_at"], unique=False
    )
    op.create_index(
        op.f("ix_nb_runtime_docker_name"), "nb_runtime", ["docker_name"], unique=True
    )
    op.create_table(
        "nb_workflow",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("wfid", sa.String(length=24), nullable=True),
        sa.Column("alias", sa.String(length=33), nullable=False),
        sa.Column("nbtask", sa.JSON(), nullable=False),
        sa.Column("schedule", sa.JSON(), nullable=False),
        sa.Column("enabled", sa.Boolean(), nullable=False),
        sa.Column(
            "created_at", sa.DateTime(), server_default=sa.text("now()"), nullable=False
        ),
        sa.Column(
            "updated_at", sa.DateTime(), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column("project_id", sa.String(length=16), nullable=True),
        sa.ForeignKeyConstraint(
            ["project_id"], ["nb_project.projectid"], ondelete="SET NULL"
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("alias", "project_id", name="_nb_workflow__project_alias"),
    )
    op.create_index(
        op.f("ix_nb_workflow_alias"), "nb_workflow", ["alias"], unique=False
    )
    op.create_index(op.f("ix_nb_workflow_wfid"), "nb_workflow", ["wfid"], unique=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_nb_workflow_wfid"), table_name="nb_workflow")
    op.drop_index(op.f("ix_nb_workflow_alias"), table_name="nb_workflow")
    op.drop_table("nb_workflow")
    op.drop_index(op.f("ix_nb_runtime_docker_name"), table_name="nb_runtime")
    op.drop_index(op.f("ix_nb_runtime_created_at"), table_name="nb_runtime")
    op.drop_table("nb_runtime")
    op.drop_table("nb_projects_users")
    op.drop_index(op.f("ix_nb_history_status"), table_name="nb_history")
    op.drop_index(op.f("ix_nb_history_created_at"), table_name="nb_history")
    op.drop_table("nb_history")
    op.drop_index(op.f("ix_nb_project_projectid"), table_name="nb_project")
    op.drop_table("nb_project")
    op.drop_index(op.f("ix_nb_auth_user_username"), table_name="nb_auth_user")
    op.drop_table("nb_auth_user")
    # ### end Alembic commands ###



from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class KeyPairs(BaseModel):
    public: str
    private: str


class JWTConfig(BaseModel):
    alg: str
    exp_min: int = 30
    keys: Optional[KeyPairs] = None
    secret: Optional[str] = None
    issuer: Optional[str] = None
    audience: Optional[str] = None
    requires_claims: List[str] = ["exp"]
    ttl_refresh_token: int = 3600 * 168  # 7 days


class UserLogin(BaseModel):
    username: str
    password: str


class JWTResponse(BaseModel):
    access_token: str
    refresh_token: Optional[str]


class TokenCreds(BaseModel):
    access_token: str
    refresh_token: Optional[str] = None

    class Config:
        extra = "forbid"



import importlib
import os

from labfunctions.types import ServerSettings

from . import load_server

# singleton pattern
# https://github.com/samuelcolvin/pydantic/issues/586
settings = load_server()



from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.types import BINARY, JSON, BigInteger


@compiles(JSON, "postgresql")
def compile_jsonb_postgres(type_, compile, **kw):
    return "JSONB"


@compiles(BINARY, "postgresql")
def compile_bytea_postgres(type_, compiler, **kw):
    return "BYTEA"


@compiles(BigInteger, "sqlite")
def compile_integer_sqlite(type_, compiler, **kw):
    return "Integer"


Base = declarative_base()


def create_all(engine):
    Base.metadata.create_all(bind=engine)


def drop_all(engine):
    Base.metadata.drop_all(bind=engine)


class MutableList(Mutable, list):
    """This allow modify an ARRAY custom type
    From https://gist.github.com/kirang89/22d111737af0fca251e3

    """

    def __setitem__(self, key, value):
        list.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        list.__delitem__(self, key)
        self.changed()

    def append(self, value):
        list.append(self, value)
        self.changed()

    def pop(self, index=0):
        value = list.pop(self, index)
        self.changed()
        return value

    @classmethod
    def coerce(cls, key, value):
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)
            return Mutable.coerce(key, value)
        else:
            return value



import importlib.util
import os

from jinja2 import Environment, FileSystemLoader, select_autoescape


def get_package_dir(pkg):
    spec = importlib.util.find_spec(pkg)
    return spec.submodule_search_locations[0]


env = Environment(
    loader=FileSystemLoader(f"{get_package_dir('labfunctions')}/conf/templates"),
    autoescape=select_autoescape(),
)


def render(filename, *args, **kwargs):
    tpl = env.get_template(filename)
    return tpl.render(*args, **kwargs)


def render_to_file(template, dst, *args, **kwargs):
    text = render(template, *args, **kwargs)
    with open(dst, "w", encoding="utf-8") as f:
        f.write(text)
    return text



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

from labfunctions.types.security import JWTResponse


class TokenStoreSpec(ABC):
    @abstractmethod
    async def put(self, key: str, value: str, ttl: Optional[int] = None) -> bool:
        pass

    @abstractmethod
    async def get(self, key: str) -> Union[str, None]:
        pass

    @abstractmethod
    async def delete(self, key: str):
        pass

    # @abstractmethod
    # async def validate(self, token: str, user: str) -> bool:
    #     pass

    @staticmethod
    @abstractmethod
    def generate(sign: Optional[str] = None) -> str:
        pass


class AuthSpec(ABC):
    @abstractmethod
    def encode(self, payload: Dict[str, Any], exp=None, iss=None, aud=None):
        pass

    @abstractmethod
    def decode(
        self, encoded, verify_signature=True, verify_exp=True, iss=None, aud=None
    ) -> Dict[str, Any]:
        pass

    @abstractmethod
    def validate(
        self,
        token: str,
        required_scopes: Optional[List[str]],
        require_all=True,
        iss=None,
        aud=None,
    ) -> Dict[str, Any]:
        pass

    @abstractmethod
    async def refresh_token(self, access_token, refresh_token) -> JWTResponse:
        pass

    @abstractmethod
    async def store_refresh_token(self, username: str) -> str:
        pass



import os
import secrets
from dataclasses import dataclass
from typing import Optional

from rich.console import Console

from labfunctions.commands import shell
from labfunctions.utils import get_free_port

console = Console()


@dataclass
class JupyterOpts:
    port: Optional[str] = None
    token: Optional[str] = None
    base_url: str = "/"
    addr: str = "127.0.0.1"


def check_jupyter_installed() -> bool:
    installed = False
    try:
        shell("jupyter --help", silent=True)
        installed = True
    except FileNotFoundError:
        pass
    return installed


def jupyter_exec(opts: JupyterOpts):

    port = opts.port or get_free_port(opts.addr)
    token = opts.token or secrets.token_urlsafe(50)
    env = os.environ
    env.update({"JUPYTER_TOKEN": token})
    is_local = os.environ.get("NB_LOCAL")
    installed = check_jupyter_installed()
    if not installed and not is_local:
        shell("pip install jupyterlab")
    elif not installed and is_local:
        console.log("[red bold](x) Jupyter lab is not installed[/]")

    console.print(
        f"=> Running jupyter into: [magenta bold]{opts.addr}:{port}{opts.base_url}[/]"
    )
    if is_local:
        console.print(f"Access token: {token}")

    cmd = (
        f"jupyter lab --ip={opts.addr}  --ServerApp.port={port} "
        f"--ServerApp.base_url={opts.base_url} "
        f"--no-browser "
        f"--ServerApp.shutdown_no_activity_timeout=15"
    )

    output = shell(cmd, env=env, silent=True)



import os
from datetime import datetime
from typing import List

import redis
from rq import Connection, Worker


class NBWorker(Worker):
    """Extensions of the default Worker class to set ip_address"""

    def set_ip_address(self, ip_address):
        self.ip_address = ip_address

    def inactive_time(self):
        """it calculates in seconds, the total inactive time of a worker"""
        working = self.total_working_time
        now = datetime.utcnow()
        birth_elapsed = (now - self.birth_date).seconds
        return birth_elapsed - working


def start_worker(redis_dsn, queues: List[str], ip_address: str, name: str):

    rdb = redis.from_url(redis_dsn)
    pid = os.getpid()

    with Connection(connection=rdb):
        print(f"Running in {pid} with ip {ip_address}", pid)
        # qs = sys.argv[1:] or ['default']
        w = NBWorker(queues, name=name)
        w.set_ip_address(ip_address)
        w.work()



import json
import logging
from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from labfunctions import defaults, errors, secrets, types
from labfunctions.errors.client import ProjectCreateError, ProjectUploadError
from labfunctions.errors.runtimes import RuntimeCreationError
from labfunctions.utils import binary_file_reader, parse_var_line

from .base import BaseClient
from .utils import store_private_key


class ProjectsClient(BaseClient):
    """Is to be used as cli client because it has side effects on local disk"""

    def projects_create(
        self,
        name: str,
        desc: Optional[str] = None,
        repository: Optional[str] = None,
        store_key: bool = True,
    ) -> Union[types.projects.ProjectCreated, None]:
        """
        A staless project creator it don't use any internal state of the client
        instead, all values should be provided in this method for a project to be
        created
        """

        pkey = secrets.generate_private_key()
        pq = types.ProjectReq(
            name=name,
            private_key=pkey,
            description=desc,
            repository=repository,
        )
        r = self._http.post(
            f"/projects",
            json=asdict(pq),
        )
        created = False
        if r.status_code == 201:
            pd = types.ProjectData(**r.json())
            pc = types.projects.ProjectCreated(pd=pd, private_key=pkey)
            if store_key:
                store_private_key(pkey, pd.projectid)
            return pc
        if r.status_code != 200:
            raise ProjectCreateError(name)
        return None

    def projects_get(self) -> Union[types.ProjectData, None]:
        r = self._http.get(f"/projects/{self.projectid}")
        if r.status_code == 200:
            return types.ProjectData(**r.json())
        return None

    def projects_list(self) -> List[types.ProjectData]:
        r = self._http.get(f"/projects")
        if r.status_code == 200:
            data = r.json()
            return [types.ProjectData(**d) for d in data]
        return []

    def projects_upload(self, zfile: types.ProjectBundleFile):
        url = (
            f"/projects/{self.projectid}"
            f"/_upload?runtime={zfile.runtime_name}&version={zfile.version}"
        )

        r = self._http.post(
            url,
            content=binary_file_reader(zfile.filepath),
        )
        if r.status_code != 201 and r.status_code != 204:
            raise ProjectUploadError(self.projectid)

    def projects_build(self, spec: types.RuntimeSpec, version: str) -> Union[str, None]:
        rsp = self._http.post(
            f"/projects/{self.projectid}/_build?version={version}", json=spec.dict()
        )
        if rsp.status_code == 202:
            return rsp.json()["execid"]
        return None

    def projects_create_agent(self) -> Union[str, None]:
        r = self._http.post(f"/projects/{self.projectid}/agent")

        if r.status_code == 201:
            return r.json()["msg"]

        return None

    def projects_agent_token(
        self, agentname: Optional[str] = None, projectid: Optional[str] = None
    ) -> Union[types.user.AgentJWTResponse, None]:
        """
        If an agentname is given then it will ask for that agent, if not, then it will
        get the token for the last agent created for this project.
        """
        prj = self.projectid
        if projectid:
            prj = projectid

        url = f"/projects/{prj}/agent/_token"
        if agentname:
            url = f"/projects/{prj}/agent/{agentname}/_token"

        r = self._http.post(url)

        if r.status_code == 200:
            return types.user.AgentJWTResponse(**r.json())

        return None

    def projects_agent_list(self) -> List[str]:
        """
        If an agentname is given then it will ask for that agent, if not, then it will
        get the token for the last agent created for this project.
        """

        url = f"/projects/{self.projectid}/agent"
        r = self._http.get(url)

        if r.status_code == 200:
            return r.json()
        return []

    def project_agent_delete(self, agentname) -> bool:

        url = f"/projects/{self.projectid}/agent/{agentname}"
        r = self._http.delete(url)
        if r.status_code == 200:
            return True
        return False

    def projects_private_key(self, store_key=False) -> Union[str, None]:
        """
        Gets private key to be shared to the docker container of a
        workflow task
        """
        r = self._http.get(f"/projects/{self.projectid}/_private_key")

        if r.status_code == 200:
            key = r.json()["private_key"]
            if store_key:
                store_private_key(key, self.projectid)
            return key
        return None

    def runtimes_get_all(self, lt=5) -> List[types.RuntimeData]:
        data = self._http.get(f"/runtimes/{self.projectid}?lt={lt}")
        runtimes = [types.RuntimeData(**dict_) for dict_ in data.json()]
        return runtimes

    def runtime_create(self, req: types.RuntimeReq):

        rsp = self._http.post(f"/runtimes/{self.projectid}", json=req.dict())
        if rsp.status_code != 201 and rsp.status_code != 200:
            raise RuntimeCreationError(
                docker_name=req.docker_name, projectid=self.projectid
            )

    def runtime_delete(self, rid):
        rsp = self._http.delete(f"/runtimes/{self.projectid}/{rid}")
        if rsp.status_code != 200:
            raise RuntimeCreationError(docker_name=rid, projectid=self.projectid)



import logging
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union
from zipfile import ZipFile

from labfunctions import defaults, secrets
from labfunctions.conf.jtemplates import get_package_dir, render_to_file
from labfunctions.errors import CommandExecutionException
from labfunctions.types.runtimes import ProjectBundleFile, RuntimeSpec
from labfunctions.utils import execute_cmd

from .utils import git_last_tag, git_short_head_id

logger = logging.getLogger(__name__)


def get_secrets_filepath(working_area: Path) -> Path:

    return working_area / defaults.CLIENT_TMP_FOLDER / defaults.SECRETS_FILENAME


def write_secrets(working_area, private_key, nbvars_dict) -> Path:
    _vars = secrets.encrypt_nbvars(private_key, nbvars_dict)
    newline = "\n"
    encoded_vars = f'{newline.join(f"{key}={value}" for key, value in _vars.items())}'
    outfile = get_secrets_filepath(working_area)
    with open(outfile, "w") as f:
        f.write(encoded_vars)

    return outfile


def zip_git_stash(
    root, working_area, runtime_name, prefix_folder=defaults.ZIP_GIT_PREFIX
) -> Union[ProjectBundleFile, None]:
    """Zip the actual folder state
    using git stash, this should be used only when testing or developing
    """
    # project = str(root).rsplit("/", maxsplit=1)[-1]

    secrets_file = get_secrets_filepath(working_area)

    filename = f"{runtime_name}.stash.zip"

    output_file = f"{str(working_area)}/{defaults.CLIENT_TMP_FOLDER}/{filename}"
    stash_id = execute_cmd("git stash create")
    if not stash_id:
        return None
    cmd = (
        f"git archive --prefix={prefix_folder} "
        f"--add-file {secrets_file} "
        f"-o {output_file} {stash_id} "
    )

    execute_cmd(cmd)

    return ProjectBundleFile(
        filepath=output_file,
        filename=filename,
        current=False,
        stash=True,
        version="stash",
        runtime_name=runtime_name,
    )


def zip_git_head(
    root, working_area, runtime_name, prefix_folder=defaults.ZIP_GIT_PREFIX
) -> ProjectBundleFile:
    """Zip the head of the repository.
    Not commited files wouldn't included in this zip file
    """
    # project = str(root).rsplit("/", maxsplit=1)[-1]

    secrets_file = get_secrets_filepath(working_area)

    tagname = None
    try:
        tagname = git_last_tag()
    except CommandExecutionException:
        logger.warning(
            "Any tag were found in the git repository, the last commit id will be used instad"
        )
    if not tagname:
        tagname = git_short_head_id()

    filename = f"{runtime_name}.{tagname}.zip"

    output_file = f"{str(working_area)}/{defaults.CLIENT_TMP_FOLDER}/{filename}"
    execute_cmd(
        f"git archive --prefix={prefix_folder} "
        f"--add-file {secrets_file} "
        f"-o {output_file} HEAD "
    )
    return ProjectBundleFile(
        runtime_name=runtime_name,
        filepath=output_file,
        filename=filename,
        version=tagname.lower(),
        commit=tagname,
    )


def zip_current(
    root, working_area, runtime_name, prefix_folder=defaults.ZIP_GIT_PREFIX
) -> ProjectBundleFile:

    filename = f"{runtime_name}.current.zip"

    output_file = f"{str(working_area)}/{defaults.CLIENT_TMP_FOLDER}/{filename}"

    secrets_file = get_secrets_filepath(working_area)
    dst = root / ".secrets"
    if dst.is_file():
        dst.write_bytes(secrets_file.read_bytes())

    with ZipFile(output_file, "w") as z:
        for i in Path(".").glob("**/*"):
            if (
                not str(i).startswith(".venv")
                and not str(i).startswith(".git")
                and not str(i).startswith(".tox")
                and not str(i).startswith(defaults.CLIENT_TMP_FOLDER)
            ):
                if prefix_folder:
                    z.write(i, (prefix_folder / i))
                else:
                    z.write(i)

    dst.unlink(missing_ok=True)
    return ProjectBundleFile(
        runtime_name=runtime_name,
        filepath=output_file,
        filename=filename,
        version="current",
        current=True,
    )


def bundle_project(
    working_area: Union[str, Path],
    spec: RuntimeSpec,
    privkey=None,
    stash=False,
    current=False,
) -> ProjectBundleFile:
    """
    It's in charge of bundle all the files needed to build a runtime.
    This bundle could be uploaded to the server or used to build a local docker image.

    Right now AGENT_TOKEN and AGENT_REFRESH_TOKEN are injected dinamically
    generating the keys on the server and puting it encrypted in the .secrets's file

    This is a little magical from the point of view of the user, but the intention is the
    simplification of the overall process, and we try to avoid that the user handle
    sensible information manually

    Same approach is taken with the private key to sign secrets. In the future this could
    change or at least we will provide to the user with the right mechanisms
    and documentation so that they can handle it manually if they want.
    """

    # secrets.nbvars["AGENT_TOKEN"] = creds.access_token
    # secrets.nbvars["AGENT_REFRESH_TOKEN"] = creds.refresh_token
    # checks if AGENT_TOKEN and REFERSH exist?

    root = Path(os.getcwd())
    wa = Path(working_area)

    if not (root / spec.container.requirements).is_file():
        raise KeyError("A requirements file is missing")

    (wa / defaults.CLIENT_TMP_FOLDER).mkdir(parents=True, exist_ok=True)

    nbvars = secrets.load(str(root))

    if privkey:
        secrets_file = write_secrets(wa, privkey, nbvars)

    zfile = None

    if stash and not current:
        zfile = zip_git_stash(root, wa, spec.name)
        if not zfile:
            raise TypeError(
                "There isn't changes in the git repository to perform a "
                "STASH zip file. For untracked files you should add to "
                "the stash the changes, perform: git add ."
            )
    elif current and not stash:
        zfile = zip_current(root, wa, spec.name)
    elif not current and not stash:
        zfile = zip_git_head(root, wa, spec.name)
    else:
        raise AttributeError("Bad option: current and stash are different options")

    if privkey:
        Path(secrets_file).unlink(missing_ok=True)

    return zfile



from datetime import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel

from labfunctions import defaults


class DockerGPUSpec(BaseModel):
    cuda: str = "11.6"
    cudnn: str = "8.4.0.27-1+cuda11.6"
    cudnn_major_version = "8"
    nvidia_gpg_key = defaults.NVIDIA_GPG_KEY
    nvidia_gpg_version = defaults.NVIDIA_GPG_VERSION


class DockerSpec(BaseModel):
    image: str
    maintainer: str
    build_packages: Optional[str] = None
    final_packages: Optional[str] = None
    user: Optional[Dict[str, int]] = None
    base_template: str = "Dockerfile.default"
    requirements: str = "requirements.txt"
    gpu: Optional[DockerGPUSpec] = None
    extra: Optional[Dict[str, Any]] = None


class RuntimeSpec(BaseModel):
    name: str
    container: DockerSpec
    description: Optional[str] = None
    gpu_support: bool = False
    version: Optional[str] = None
    registry: Optional[str] = None


class RuntimeReq(BaseModel):
    runtime_name: str
    docker_name: str
    spec: RuntimeSpec
    project_id: str
    version: str
    registry: Optional[str]


class RuntimeData(BaseModel):
    """
    docker_name should be nbworkflows/[projectid]-[runtime_name]:[version]
    """

    runtimeid: str
    runtime_name: str
    docker_name: str
    spec: RuntimeSpec
    project_id: str
    version: str
    created_at: Optional[str] = None
    registry: Optional[str] = None
    id: Optional[int] = None


class ProjectBundleFile(BaseModel):
    runtime_name: str
    version: str
    filepath: str
    filename: str
    commit: Optional[str]
    stash: Optional[bool] = False
    current: Optional[bool] = False
    format_type: str = "zip"


class BuildCtx(BaseModel):
    """
    It's is in charg of bring all the information needed to build
    docker containers.

    It has two entries: projects_bp and cli

    :param projectid: is needed to register the runtime built
    :param spec: Spec of the runtime to built
    :param docker_name: name of the docker
    :param version: version of the runtime, used to tag the docker image
    :param dockerfile: Name of the Dockerfile, by default is: "Dockerfile.default"
    :param zip_name: zip name: CURRENT.zip
    :param download_zip: path to the zip file, this shouldn't include
    nor bucket nor url
    :param execid: random id to register this task
    :param project_store_class: which type of storage use to download the bundle
    :param project_store_bucket: bucket to find the bundle file.

    :param registry: registry to push the docker image built
    """

    projectid: str  # ok
    spec: RuntimeSpec
    docker_name: str
    version: str
    dockerfile: str
    zip_name: str
    download_zip: str
    execid: str
    project_store_class: str
    project_store_bucket: str
    registry: Optional[str] = None



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from sqlalchemy import delete, select, update
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import selectinload

from labfunctions import context as ctx
from labfunctions import defaults, errors
from labfunctions.errors.generics import WorkflowRegisterError
from labfunctions.hashes import generate_random
from labfunctions.managers import projects_mg, runtimes_mg
from labfunctions.models import WorkflowModel
from labfunctions.types import (
    ExecutionNBTask,
    NBTask,
    ProjectData,
    ScheduleData,
    WorkflowData,
    WorkflowDataWeb,
    WorkflowsList,
)
from labfunctions.utils import get_version

WFDATA_RULES = ("-id", "-project", "-project_id", "-created_at", "-updated_at")


def _update(wfid: str, projectid: str, wfd: WorkflowDataWeb):

    task_dict = wfd.nbtask.dict()
    schedule = None
    if wfd.schedule:
        schedule = wfd.schedule.dict()

    wm_table = WorkflowModel.__table__
    stmt = (
        update(wm_table)
        .where(wm_table.c.wfid == wfid)
        .values(
            nbtask=task_dict,
            schedule=schedule,
            alias=wfd.alias,
            enabled=wfd.enabled,
            updated_at=datetime.utcnow(),
        )
    )
    return stmt


def select_workflow():
    stmt = select(WorkflowModel).options(selectinload(WorkflowModel.project))
    return stmt


def generate_wfid(size=defaults.WFID_LEN) -> str:
    """wfid refers to the workflow id, this is only defined once, when the
    workflow is created, and should to be unique."""
    return generate_random(size)


def get_job_from_db(session, wfid) -> Union[WorkflowModel, None]:
    stmt = select(WorkflowModel).where(WorkflowModel.wfid == wfid)
    result = session.execute(stmt)
    row = result.scalar()

    if row:
        return row
    return None


async def get_by_wfid(session, wfid) -> WorkflowData:
    stmt = select_workflow().where(WorkflowModel.wfid == wfid)
    result = await session.execute(stmt)
    row = result.scalar()
    if row:
        return WorkflowData(**row.to_dict(rules=WFDATA_RULES))
    return None


async def get_by_wfid_prj(session, projectid, wfid) -> WorkflowData:
    stmt = (
        select_workflow()
        .where(WorkflowModel.wfid == wfid)
        .where(WorkflowModel.project_id == projectid)
    )
    result = await session.execute(stmt)
    row = result.scalar()
    if row:
        return WorkflowData(**row.to_dict(rules=WFDATA_RULES))
    return None


def get_by_prj_and_wfid_sync(session, projectid, wfid) -> Union[WorkflowModel, None]:
    stmt = (
        select_workflow()
        .where(WorkflowModel.project_id == projectid)
        .where(WorkflowModel.wfid == wfid)
    )
    result = session.execute(stmt)
    row = result.scalar()
    if row:
        return row
    return None


async def get_all(session, project_id=None) -> List[WorkflowData]:
    stmt = select_workflow()
    if project_id:
        stmt = stmt.where(WorkflowModel.project_id == project_id)

    result = await session.execute(stmt)
    rows = result.scalars()

    wfs = []
    for r in rows:
        dict_ = r.to_dict(rules=WFDATA_RULES)
        wfs.append(WorkflowData(**dict_))

    return wfs


async def get_by_alias(session, alias) -> Union[WorkflowModel, None]:
    stmt = select_workflow().where(WorkflowModel.alias == alias).limit(1)
    result = await session.execute(stmt)
    row = result.scalar_one_or_none()
    return row


async def register(session, projectid: str, wfd: WorkflowDataWeb, update=False) -> str:
    """Register workflows"""
    wfid = generate_wfid()

    if update:
        stmt = _update(wfd.wfid, projectid, wfd)
        await session.execute(stmt)
        return wfd.wfid

    data_dict = wfd.dict()
    data_dict["wfid"] = wfid
    schedule = None
    if wfd.schedule:
        schedule = wfd.schedule.dict()

    obj = WorkflowModel(
        wfid=wfid,
        alias=wfd.alias,
        nbtask=wfd.nbtask.dict(),
        schedule=schedule,
        project_id=projectid,
        enabled=wfd.enabled,
    )
    session.add(obj)
    try:
        await session.flush()
    except IntegrityError as e:
        await session.rollback()
        raise WorkflowRegisterError(projectid, wfd.alias, e)
    return wfid


async def delete_wf(session, project_id, wfid):
    stmt = (
        delete(WorkflowModel)
        .where(WorkflowModel.project_id == project_id)
        .where(WorkflowModel.wfid == wfid)
    )
    await session.execute(stmt)


async def prepare_notebook_job_async(
    session, projectid: str, wfid: str, execid: str
) -> ExecutionNBTask:
    """It prepares the task execution of the notebook"""
    wm = await get_by_wfid_prj(session, projectid, wfid)
    if wm and wm.enabled:
        pm = await projects_mg.get_by_projectid_model(session, projectid)
        task = NBTask(**wm.task)

        pd = ProjectData.from_orm(pm)
        pd.username = pm.user.username

        exec_notebook_ctx = ctx.create_notebook_ctx(pd, task, execid)
        return exec_notebook_ctx
    elif not wm.enabled:
        raise errors.WorkflowDisabled(projectid, wfid)

    raise errors.WorkflowNotFound(projectid, wfid)



from . import user
from .client import WorkflowsFile
from .config import ClientSettings, ServerSettings
from .core import (
    ExecutionNBTask,
    ExecutionResult,
    HistoryLastResponse,
    HistoryRequest,
    HistoryResult,
    NBTask,
    ScheduleData,
    SimpleExecCtx,
    WorkflowData,
    WorkflowDataWeb,
    WorkflowsList,
)
from .projects import ProjectData, ProjectReq
from .runtimes import ProjectBundleFile, RuntimeData, RuntimeReq, RuntimeSpec
from .security import TokenCreds



import sys
from datetime import datetime
from typing import Optional

import click
import redis
from rich import print_json
from rich.prompt import Confirm
from rich.table import Table

from labfunctions import defaults
from labfunctions.cluster import create_cluster_control
from labfunctions.conf.server_settings import settings
from labfunctions.control_plane.register import AgentRegister
from labfunctions.errors.cluster import ClusterSpecNotFound
from labfunctions.utils import format_seconds

from .utils import console, progress


@click.group(name="cluster")
def clustercli():
    """
    Execute workflows or notebooks, locally or remote
    """
    pass


@clustercli.command(name="create-machine")
@click.option(
    "--from-file",
    "-f",
    default=settings.CLUSTER_SPEC,
    help="yaml file with the configuration",
)
@click.option(
    "--deploy",
    "-D",
    is_flag=True,
    default=False,
    help="If used, then a agent will be deployed",
)
@click.option("--cluster", "-C", default=None, help="Cluster where it will run")
@click.option("--use-public", "-P", is_flag=True, default=False, help="Use public ip")
@click.option("--deploy-local", "-L", is_flag=True, default=False, help="Run locally")
def create_machinecli(from_file, cluster, deploy, use_public, deploy_local):
    """It will create a new machine in a provider choosen, inside of a cluster"""
    try:
        cc = create_cluster_control(from_file, settings.RQ_REDIS, cluster)
    except ClusterSpecNotFound as e:
        console.print(f"[bold red]{e}[/]")
        sys.exit(-1)

    instance = cc.create_instance(settings, deploy, use_public, deploy_local)
    print_json(data=instance.dict())


@clustercli.command(name="destroy-machine")
@click.option(
    "--from-file",
    "-f",
    default=settings.CLUSTER_SPEC,
    help="yaml file with the configuration",
)
@click.option("--cluster", "-C", default=None, help="Cluster where it will run")
@click.argument("machine-name")
def destroy_machinecli(from_file, machine_name, cluster):
    """Destroy a machine by name"""

    try:
        cc = create_cluster_control(from_file, settings.RQ_REDIS, cluster)
        console.print(
            f"=> Destroying [magenta bold]{machine_name}[/] in "
            f"[cyan bold]{cc.cluster_name}[/]"
        )
    except ClusterSpecNotFound as e:
        console.print(f"[bold red]{e}[/]")
        sys.exit(-1)

    cc.destroy_instance(machine_name)
    console.print(f"[green bold]{machine_name}[/] [green]destroyed[/]")


@clustercli.command(name="list-machines")
@click.option(
    "--from-file",
    "-f",
    default=settings.CLUSTER_SPEC,
    help="yaml file with the configuration",
)
@click.option("--tags", "-T", default="nbworkflows", help="Tags separated by comma")
@click.option("--cluster", "-C", default=None, help="Cluster where it will run")
@click.option(
    "--filter-by-cluster",
    "-F",
    is_flag=True,
    default=True,
    help="Filter nodes by cluster",
)
def list_machinescli(from_file, tags, cluster, filter_by_cluster):
    """List machines in a cluster"""
    try:
        cc = create_cluster_control(from_file, settings.RQ_REDIS, cluster)
    except ClusterSpecNotFound as e:
        console.print(f"[bold red]{e}[/]")
        sys.exit(-1)

    driver = cc.inventory.get_provider(cc.spec.provider)
    with progress:
        task = progress.add_task(
            f"Getting machines in cluster {cc.cluster_name}", start=False, total=1
        )
        nodes = driver.list_machines(location=cc.spec.location, tags=tags.split(","))
        progress.advance(task)
    title = "Machines"
    if filter_by_cluster:
        filtered = [n for n in nodes if n.labels["cluster"] == cc.cluster_name]
        title = f"Machines for cluster {cc.cluster_name}"
        nodes = filtered
    table = Table(title=title)
    table.add_column("Name", justify="left", style="red")
    table.add_column("Location", justify="center", style="cyan")
    table.add_column("Private", justify="center", style="cyan")
    table.add_column("Public", style="cyan", justify="center")
    table.add_column("Cluster", style="red", justify="right")
    for n in nodes:
        table.add_row(
            n.machine_name,
            n.location,
            n.private_ips[0],
            n.public_ips[0],
            n.labels["cluster"],
        )
    console.print(table)


@clustercli.command(name="up")
@click.option(
    "--from-file",
    "-f",
    default=settings.CLUSTER_SPEC,
    help="yaml file with the configuration",
)
@click.option("--cluster", "-C", default=None, help="Cluster where it will run")
@click.option("--use-public", "-P", is_flag=True, default=False, help="Use public ip")
@click.option("--deploy-local", "-L", is_flag=True, default=False, help="Run locally")
@click.option("--do-deploy", "-D", is_flag=True, default=True, help="Deploy agent")
def upcli(from_file, use_public, deploy_local, cluster, do_deploy):
    """Create a cluster"""
    try:
        cc = create_cluster_control(from_file, settings.RQ_REDIS, cluster)
    except ClusterSpecNotFound as e:
        console.print(f"[bold red]{e}[/]")
        sys.exit(-1)

    with progress:
        progress.add_task(f"Starting {cc.cluster_name}...", start=False)
        cc.scale(
            settings,
            use_public=use_public,
            deploy_local=deploy_local,
            do_deploy=do_deploy,
        )

    console.print(f"=> [green bold]Cluster {cc.cluster_name} started[/]")


@clustercli.command(name="destroy")
@click.option(
    "--from-file",
    "-f",
    default=settings.CLUSTER_SPEC,
    help="yaml file with the configuration",
)
@click.option("--cluster", "-C", default=None, help="Cluster where it will run")
@click.option(
    "--hard", is_flag=True, default=False, help="if true it destroy machines directly"
)
def destroycli(from_file, cluster, hard):
    """Destroy a cluster"""
    try:
        cc = create_cluster_control(from_file, settings.RQ_REDIS, cluster)
    except ClusterSpecNotFound as e:
        console.print(f"[bold red]{e}[/]")
        sys.exit(-1)

    with progress:
        progress.add_task(f"Destroying cluster {cc.cluster_name}")
        if hard:
            nodes = cc.provider.list_machines(
                location=cc.spec.location, tags=[defaults.CLOUD_TAG]
            )
            filtered = [n for n in nodes if n.labels["cluster"] == cluster]
            for n in filtered:
                progress.add_task(f"Destroying {n.machine_name}")
                cc.provider.destroy_machine(n)
        else:
            agents = cc.register.list_agents(cluster)
            for agt in agents:
                progress.add_task(f"Destroying {agt}")
                cc.destroy_instance(agt)
    console.print(
        f"[bold]Cluster [magenta bold]{cc.cluster_name}[/magenta bold] destroyed[/]"
    )


@clustercli.command(name="list-agents")
@click.option(
    "--from-file",
    "-f",
    default=settings.CLUSTER_SPEC,
    help="yaml file with the configuration",
)
@click.option("--provider", "-p", default="gce", help="Cloud to run")
@click.option("--cluster", "-C", default=None, help="Cluster")
def list_agentscli(from_file, provider, cluster):
    """List agents by cluster"""

    rdb = redis.from_url(settings.RQ_REDIS, decode_responses=True)
    ar = AgentRegister(rdb, cluster)

    agents = ar.list_agents(from_cluster=cluster)
    title = "Agents"
    if cluster:
        title = f"Agents for {cluster}"
    table = Table(title=title)
    table.add_column("Name", justify="left", style="cyan")
    table.add_column("Queues", justify="center", style="cyan")
    table.add_column("Cluster", justify="center", style="red")
    table.add_column("Started", style="red", justify="right")

    now = datetime.utcnow().timestamp()
    for agt in agents:
        _agt = ar.get(agt)
        elapsed = now - _agt.birthday
        table.add_row(
            _agt.name, ",".join(_agt.qnames), _agt.cluster, format_seconds(elapsed)
        )
    console.print(table)


@clustercli.command(name="list-clusters")
@click.option(
    "--from-file",
    "-f",
    default=settings.CLUSTER_SPEC,
    help="yaml file with the configuration",
)
@click.option("--cluster", "-C", default=None, help="Cluster")
def list_clusters(from_file, cluster):
    """It will run the task inside a docker container, it exist only as a tester"""

    rdb = redis.from_url(settings.RQ_REDIS, decode_responses=True)
    ar = AgentRegister(rdb, cluster)

    clusters = ar.list_clusters()
    title = "Clusters"
    table = Table(title=title)
    table.add_column("Name", justify="left", style="cyan")
    for c in clusters:
        table.add_row(c)
    console.print(table)


@clustercli.command(name="catalog")
@click.option(
    "--from-file",
    "-f",
    default=settings.CLUSTER_SPEC,
    help="yaml file with the configuration",
)
@click.option("--provider", "-p", default=None, help="From a provider")
def catalogcli(from_file, provider):
    """List machine types"""

    try:
        cc = create_cluster_control(from_file, settings.RQ_REDIS)
    except ClusterSpecNotFound as e:
        console.print(f"[bold red]{e}[/]")
        sys.exit(-1)

    inventory = cc.inventory

    title = "Machines types"
    table = Table(title=title)
    table.add_column("name", justify="left", style="cyan")
    table.add_column("provider", justify="center", style="blue")
    table.add_column("location", justify="center", style="blue")
    table.add_column("size", justify="right", style="blue")
    for _, m in inventory.machines.items():
        if provider:
            if m.provider == provider:
                table.add_row(m.name, m.provider, m.location, m.machine_type.size)
        else:
            table.add_row(m.name, m.provider, m.location, m.machine_type.size)
    console.print(table)



# import tempfile
import json
import logging
import time
from dataclasses import asdict, dataclass
from datetime import datetime
from functools import wraps
from typing import Any, Dict, Optional

import cloudpickle
import httpx

from labfunctions.conf.client_settings import settings

# from labfunctions.workflows.core import build_context
from labfunctions.types import SimpleExecCtx

logger = logging.getLogger(__name__)

VALID_STRATEGIES = ["local", "fileserver"]


@dataclass
class CacheConfig:
    name: str
    ctx: SimpleExecCtx
    valid_for_min: int = 60
    strategy: str = "local"


def build_ctx_global(globals_dict) -> SimpleExecCtx:
    wfid = globals_dict.get("WFID")
    execid = globals_dict.get("EXECUTIONID")
    _now = datetime.utcnow().isoformat()
    now = globals_dict.get("NOW", _now)
    return SimpleExecCtx(
        wfid=wfid,
        execid=execid,
        execution_dt=now,
    )


def build_ctx(wfid, execid, now=None) -> SimpleExecCtx:
    _now = now or datetime.utcnow().isoformat()
    return SimpleExecCtx(
        wfid=wfid,
        execid=execid,
        execution_dt=_now,
    )


def _write_pickle(name, data, ctx: SimpleExecCtx):
    fpath = f"/tmp/{ctx.wfid}.{ctx.execid}.{name}.pickle"
    metapath = f"/tmp/{ctx.wfid}.{ctx.execid}.{name}.json"
    with open(fpath, "wb") as f:
        f.write(cloudpickle.dumps(data))
        logger.debug("CACHE: Wrote to %s", fpath)
    with open(metapath, "w", encoding="utf-8") as f:
        json.dump(asdict(ctx), f)


def _restore_pickle(name, ctx: SimpleExecCtx):
    fpath = f"/tmp/{ctx.wfid}.{ctx.execid}.{name}.pickle"
    metapath = f"/tmp/{ctx.wfid}.{ctx.execid}.{name}.json"
    try:
        with open(fpath, "rb") as f:
            data = cloudpickle.load(f)
            logger.debug("CACHE: Reading from %s", fpath)
        with open(metapath, "r") as f:
            data = f.read()
            meta = json.loads(data)

        return data, meta
    except EOFError:
        return None
    except FileNotFoundError:
        return None


def _write_fileserver(name, data, ctx: SimpleExecCtx):
    urlpath = f"{settings.EXT_KV_LOCAL_ROOT}/cache/{ctx.wfid}.{ctx.execid}.{name}"
    metapath = f"{settings.EXT_KV_LOCAL_ROOT}/cache/{ctx.wfid}.{ctx.execid}.{name}.json"
    blob = cloudpickle.dumps(data)
    rsp = httpx.put(urlpath, content=blob)
    rsp2 = httpx.put(metapath, json=asdict(ctx))
    if rsp.status_code > 202 and rsp2.status_code > 202:
        logger.warning("CACHE: writing to fileserver failed %s", urlpath)
    else:
        logger.debug("CACHE: wrote to fileserver %s", urlpath)


def _restore_fileserver(name, ctx: SimpleExecCtx):
    urlpath = f"{settings.EXT_KV_LOCAL_ROOT}/cache/{ctx.wfid}.{ctx.execid}.{name}"
    metapath = f"{settings.EXT_KV_LOCAL_ROOT}/cache/{ctx.wfid}.{ctx.execid}.{name}.json"
    data = None
    meta = None

    try:
        rsp = httpx.get(urlpath)
        rsp2 = httpx.get(metapath)
        if rsp.status_code == 200 and rsp2.status_code == 200:
            data = cloudpickle.loads(rsp.content)
            meta = rsp2.json()
            logger.debug("CACHE: Reading from fileserver %s", urlpath)
    except Exception as e:
        logger.warning(e)
    return data, meta


def is_valid_date(cache_dt: str, valid_for_min: int) -> bool:
    dt = datetime.fromisoformat(cache_dt)
    now = datetime.utcnow()
    elapsed = round((now - dt).seconds / 60)
    if elapsed > valid_for_min:
        return False
    return True


def cache_manager_write(data, conf: CacheConfig):
    if conf.strategy == "local" and conf.ctx.wfid:
        _write_pickle(conf.name, data, conf.ctx)
    elif conf.strategy == "fileserver" and conf.ctx.wfid:
        _write_fileserver(conf.name, data, conf.ctx)
    else:
        logger.warning("CACHE: Invalid caching strategy %s", conf.strategy)


def cache_manager_read(conf: CacheConfig):
    data = None
    meta = None
    if conf.strategy == "local" and conf.ctx.wfid:
        data, meta = _restore_pickle(conf.name, conf.ctx)
    elif conf.strategy == "fileserver" and conf.ctx.wfid:
        data, meta = _restore_fileserver(conf.name, conf.ctx)
    else:
        logger.warning("CACHE: Invalid caching strategy %s", conf.strategy)
    return data, meta


def frozen_result(
    name=None,
    wfid=None,
    execid=None,
    valid_for_min=60,
    strategy="local",
    from_global: Optional[Dict[str, Any]] = None,
):

    if not wfid and not from_global:
        raise TypeError("wfid or global should be provided")

    if from_global:
        ctx = build_ctx_global(from_global)
    else:
        ctx = build_ctx(wfid, execid)

    cache_conf = CacheConfig(
        name=name, ctx=ctx, valid_for_min=valid_for_min, strategy=strategy
    )

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not cache_conf.name:
                cache_conf.name = func.__name__

            result, meta = cache_manager_read(cache_conf)
            valid = False
            if meta:
                valid = is_valid_date(meta["execution_dt"], valid_for_min)
            if not result or not valid:
                result = func(*args, **kwargs)
                cache_manager_write(result, cache_conf)
            return result

        return wrapper

    return decorate


if __name__ == "__main__":

    @frozen_result(wfid="test", execid="test_t", strategy="fileserver", valid_for_min=5)
    def my_func(msg):
        print(msg)
        return msg * 2

    my_func("hello world")



import os
import sys

import click

from labfunctions.conf.server_settings import settings
from labfunctions.control_plane import rqscheduler
from labfunctions.types.agent import AgentConfig
from labfunctions.utils import get_external_ip, get_hostname

hostname = get_hostname()


@click.command(name="web")
@click.option("--host", "-H", default="0.0.0.0", help="Listening Host")
@click.option("--port", "-p", default="8000", help="Listening Port")
@click.option("--workers", "-w", default=1, help="How many workers start?")
@click.option(
    "--apps",
    "-a",
    default="workflows,projects,history,events,runtimes",
    help="List of apps to be mounted as blueprints",
)
@click.option(
    "--auto-reload", "-A", default=False, is_flag=True, help="Enable Auto reload"
)
@click.option(
    "--access-log", "-L", default=False, is_flag=True, help="Enable access_log"
)
@click.option("--debug", "-D", default=False, is_flag=True, help="Enable Auto reload")
def webcli(host, port, workers, apps, auto_reload, access_log, debug):
    """Run API Web Server"""
    # pylint: disable=import-outside-toplevel
    from labfunctions.server import create_app

    list_bp = apps.split(",")
    app = create_app(settings, list_bp)
    w = int(workers)
    print("Debug mode: ", debug)
    app.run(
        host=host,
        port=int(port),
        workers=w,
        auto_reload=auto_reload,
        debug=debug,
        access_log=access_log,
    )


@click.command(name="scheduler")
@click.option("--redis", "-r", default=settings.RQ_REDIS, help="Redis full dsn")
@click.option(
    "--interval", "-i", default=60, help="How often the scheduler checks for work"
)
@click.option("--log-level", "-L", default="INFO")
def schedulercli(redis, interval, log_level):
    """Run RQ scheduler"""
    # pylint: disable=import-outside-toplevel
    rqscheduler.run(redis, interval, log_level)



import json
import logging
from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from labfunctions import defaults, errors, secrets
from labfunctions.context import create_notebook_ctx
from labfunctions.errors.generics import WorkflowRegisterClientError
from labfunctions.runtimes import local_runtime_data
from labfunctions.types import (
    ExecutionNBTask,
    ExecutionResult,
    HistoryRequest,
    HistoryResult,
    NBTask,
    ProjectData,
    ProjectReq,
    ScheduleData,
    WorkflowData,
    WorkflowDataWeb,
    WorkflowsList,
)
from labfunctions.types.workflows import WFCreateRsp, WFPushRsp
from labfunctions.utils import parse_var_line

from .base import BaseClient
from .utils import get_private_key, store_credentials_disk, store_private_key


class WorkflowsClient(BaseClient):
    """Related to workflows"""

    def workflows_create(self, wd: WorkflowDataWeb) -> WFCreateRsp:
        r = self._http.post(
            f"/workflows/{self.projectid}",
            json=wd.dict(),
        )
        code = r.status_code
        if code == 200:
            self.logger.warning(f"Workflow {wd.wfid} already exist")
        if code == 400 or code == 503:
            raise WorkflowRegisterClientError(wd.wfid, self.projectid)

        return WFCreateRsp(
            status_code=r.status_code,
            alias=wd.alias,
            msg=r.json().get("msg"),
            wfid=r.json().get("wfid"),
        )

    def workflows_update(self, wd: WorkflowDataWeb) -> WFCreateRsp:
        r = self._http.put(
            f"/workflows/{self.projectid}",
            json=wd.dict(),
        )
        code = r.status_code
        if code == 400 or code == 503:
            raise WorkflowRegisterClientError(wd.wfid, self.projectid)

        return WFCreateRsp(
            status_code=r.status_code,
            alias=wd.alias,
            msg=r.json().get("msg"),
            wfid=r.json().get("wfid"),
        )

    def workflows_push(self, refresh_workflows=True, update=False) -> WFPushRsp:
        _workflows = self.state.snapshot()
        errors = []
        created = []
        for _, wd in _workflows.workflows.items():
            try:
                if update:
                    wr = self.workflows_update(wd)
                    created.append(wr)
                else:
                    if not wd.wfid:
                        wr = self.workflows_create(wd)
                        if wr.status_code == 200:
                            self.logger.warning(f"Workflow {wd.alias} already exist")
                        elif wr.status_code == 201:
                            created.append(wr)
                            if refresh_workflows:
                                wd.wfid = wr.wfid
                                self.state.add_workflow(wd)
                        else:
                            errors.append(
                                WFCreateRsp(
                                    wfid=wd.wfid,
                                    alias=wd.alias,
                                    status_code=wr.status_code,
                                )
                            )
            except WorkflowRegisterClientError as e:
                self.logger.error(e)
                errors.append(
                    WFCreateRsp(wfid=wd.wfid, alias=wd.alias, status_code=503)
                )

        # self._workflows = _workflows
        if refresh_workflows:
            self.write()
        return WFPushRsp(created=created, errors=errors)

    def workflows_list(self) -> List[WorkflowData]:
        r = self._http.get(f"/workflows/{self.projectid}")

        return [WorkflowData(**r) for r in r.json()["rows"]]

    def workflows_get(self, wfid) -> Union[WorkflowDataWeb, None]:
        r = self._http.get(f"/workflows/{self.projectid}/{wfid}")

        if r.status_code == 200:
            return WorkflowDataWeb(**r.json())
        if r.status_code == 404:
            return None

    def workflows_delete(self, wfid) -> int:
        r = self._http.delete(f"/workflows/{self.projectid}/{wfid}")
        wd = self.state.find_by_id(wfid)
        wd.wfid = None
        if r.status_code == 200:
            # self.sync_file()
            pass
        return r.status_code

    def workflows_enqueue(self, wfid) -> str:
        r = self._http.post(f"/workflows/{self.projectid}/queue/{wfid}")
        if r.status_code == 202:
            return r.json()["execid"]

    def notebook_run(
        self,
        nb_name: str,
        params: Optional[Dict[str, Any]] = None,
        cluster=None,
        machine=None,
        runtime=None,
        version=None,
    ) -> ExecutionNBTask:

        task = NBTask(
            nb_name=nb_name,
            params=params,
            cluster=cluster,
            machine=machine,
            runtime=runtime,
            version=version,
        )
        rsp = self._http.post(
            f"/workflows/{self.projectid}/notebooks/_run", json=task.dict()
        )
        if rsp.status_code != 202:
            raise AttributeError(rsp.text)

        return ExecutionNBTask(**rsp.json())

    def build_context(
        self,
        wfid: str,
        runtime_name="default",
        runtimes_file="runtimes.yaml",
        version="latest",
    ) -> ExecutionNBTask:
        wf = self.state.find_by_id(wfid)
        rd = local_runtime_data(
            self.projectid, runtime_name, runtimes_file=runtimes_file, version=version
        )
        ctx = create_notebook_ctx(self.projectid, wf.nbtask, runtime=rd)
        return ctx



import tempfile
from typing import List, Optional

from labfunctions import defaults
from labfunctions.cluster import ssh
from labfunctions.conf.jtemplates import render_to_file
from labfunctions.types import ServerSettings
from labfunctions.types.agent import AgentRequest
from labfunctions.utils import run_sync


def _prepare_agent_cmd(
    ip_address: str,
    qnames: str,
    env_file=".env.dev.docker",
    workers=1,
):

    cmd = f"nb agent -w {workers} -i {ip_address} -q {qnames}"
    return cmd


def _prepare_docker_cmd(
    ip_address: str,
    qnames: str,
    docker_image: str,
    env_file=".env.dev.docker",
    workers=1,
    docker_version="latest",
):

    nb_agent_cmd = _prepare_agent_cmd(ip_address, qnames, env_file, workers)

    cmd = (
        f"docker run -d -v /var/run/docker.sock:/var/run/docker.sock "
        f"-e NB_SERVER=true  --env-file={env_file} "
        f"{docker_image}:{docker_version} "
        f"{nb_agent_cmd}"
    )
    return cmd


def agent(settings: ServerSettings, req: AgentRequest):
    """
    Deploy an agent into a server, it has two steps:
    render and copy a .env.docker file into the remote server
    and pull and start the agent's docker instance
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        env_file = f"{tmpdir}/.env.docker"
        render_to_file(defaults.AGENT_ENV_TPL, env_file)
        run_sync(
            ssh.scp_from_local,
            remote_addr=req.node_ip,
            remote_dir=req.agent_homedir,
            local_file=env_file,
            keys=[req.private_key_path],
        )

    agent_env_file = f"{req.agent_homedir}/.env.docker"
    addr = req.advertise_addr or req.node_ip
    cmd = _prepare_docker_cmd(
        addr,
        qnames=",".join(req.qnames),
        docker_image=req.docker_image,
        env_file=agent_env_file,
        workers=req.worker_procs,
        docker_version=req.docker_version,
    )

    result = run_sync(ssh.run_cmd, req.node_ip, cmd, keys=[req.private_key_path])
    return result



from typing import List

from sqlalchemy import delete as sqldelete
from sqlalchemy import insert as sqlinsert
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError

from labfunctions.models import MachineModel
from labfunctions.types.machine import MachineOrm, MachineType


def _insert(machine: MachineOrm):
    t = MachineModel.__table__
    stmt = sqlinsert(t).values(**machine.dict())
    return stmt


def _select_machine():
    stmt = select(MachineModel)
    return stmt


async def create(session, machine: MachineOrm) -> bool:
    stmt = _insert(machine)
    inserted = True
    try:
        r = await session.execute(stmt)
    except IntegrityError:
        inserted = False

    return inserted


async def create_or_update(session, machine: MachineOrm) -> MachineModel:
    m = MachineModel(**machine.dict())

    await session.merge(m)
    return m


def create_or_update_sync(session, machine: MachineOrm) -> MachineModel:
    m = MachineModel(**machine.dict())

    session.merge(m)
    return m


async def get_list(session, limit=1) -> List[MachineOrm]:
    stmt = _select_machine().order_by(MachineModel.created_at.desc()).limit(limit)
    rows = await session.execute(stmt)

    return [MachineOrm.from_orm(r[0]) for r in rows]


async def get_one(session, name: str) -> MachineModel:
    stmt = _select_machine().where(MachineModel.name == name).limit(1)
    rsp = await session.execute(stmt)
    return rsp.scalar_one_or_none()



import logging
from enum import Enum

# from warnings import warn


class Colors(str, Enum):  # no cov
    END = "\033[0m"
    BLUE = "\033[01;34m"
    GREEN = "\033[01;32m"
    YELLOW = "\033[01;33m"
    RED = "\033[01;31m"


srv_logger = logging.getLogger("nbwork.server")  # no cov
"""
General Server logger
"""

error_logger = logging.getLogger("nbwork.error")  # no cov
"""
Logger used for errors
"""

c_logger = logging.getLogger("nbwork.agent")  # no cov
"""
Logger used by the client
"""



from typing import Any, Dict, List, NewType, Optional, Set, Tuple, Union

from pydantic import BaseModel, BaseSettings

from labfunctions import defaults

ExtraField = NewType("ExtraField", Dict[str, Any])


class SSHKey(BaseModel):
    """
    It represents a SSHKey configuration,
    it will have the paths to public and private key
    and user associated to that key
    """

    public_path: str
    private_path: Optional[str] = None
    user: str = "op"


class MachineType(BaseModel):
    """A generic Machine type to be used in MachineModel"""

    size: str
    image: str
    vcpus: int = 1
    network: str = "default"
    extra: Optional[ExtraField] = None


class MachineGPU(BaseModel):
    """A generic representation of GPU resource"""

    name: str
    gpu_type: str
    count: int = 1
    extra: ExtraField = None


class BlockStorage(BaseModel):
    """
    A generic representation of a disk, potentially the mount_point could
    be used to identify if this will be a boot disk (needed in GCE)
    """

    name: str
    size: Union[int, str]
    location: str
    mount: str = "/mnt/disk0"
    create_if_not_exist: bool = False
    snapshot: Optional[str] = None
    image: Optional[str] = None
    permissions: Optional[List[str]] = None
    description: Optional[str] = None
    kind: Optional[str] = None
    extra: Optional[ExtraField] = None


class MachineOrm(BaseModel):
    """Entity representing the MachineModel"""

    name: str
    provider: str
    location: str  # zone
    machine_type: MachineType
    gpu: Optional[MachineGPU] = None
    volumes: List[str] = []
    desc: Optional[str] = None

    class Config:
        orm_mode = True


class MachineRequest(BaseModel):
    """
    Machine Request definition, this will be used as a request to
    create a machine

    A note about tags and labels:
    The main idea behind tags or labels in the context of NBWorkflows,
    is to differentiate machines between clusters. Maybe in the future
    another uses case could emerge.

    Some providers, like GoogleCloud, has the two options
    where the "tags" param is a list used more likely for networking/firewall rules,
    and labels is a dict that could be used for filtering or billing purposes.

    So, in the context of NBWorkflows, only should exist the concept of labels as dict.
    A workaround to address different kind of tags/labels from providers is adopting
    some conventions. For instance if the provider only allows a list of tags,
    it could be serialized in this way:

    .. code-block:
    original_labels = {"cluster": "medium-gpu", "tags": ["nbworkflows", "control_plane"] }
    provider_that_only_accepts_list =
              ["cluster:medium-gpu", "tags:nbworkflows", "tags:control_plane"]


    :param name: the machine will be created with this name, usually
    is build as `machine-type`-`random_id`
    :param size: node type in the vendor cloud parlance
    :param image: image to be used like debian or custom
    :param location: a general term, cloud providers could use zone, region or both
    :param disks: Disks to be attached to the machine creation
    :param gpu: Optional GPU resource
    :param ssh_public_cert: certificate to be added to authorized_keys
    in the remote host, this should be the string version of the certificate.
    :param ssh_user: to which user allow access.
    :param network: virtual network to configurate
    :param labels:  cluster and other properties to be used
    :param external_ip: the external IP address to use. If ‘dynamic’ (default)
    is up to the provider to asign an ip address. If ‘None’,
    then no external address will be used.
    :param internal_ip: the external IP address to use. If ‘dynamic’ (default)
    is up to the provider to asign an ip address. If ‘None’,
    then no external address will be used.
    :param extra: you should try not to use it, but is here as backup for any edge case.

    """

    name: str
    size: str
    image: str
    location: str  # zone
    internal_ip: Union[str, None] = "dynamic"
    external_ip: Union[str, None] = "dynamic"
    volumes: List[BlockStorage] = []
    gpu: Optional[MachineGPU] = None
    ssh_public_cert: Optional[str] = None
    ssh_user: Optional[str] = "op"
    network: str = "default"
    labels: Optional[Dict[str, Any]] = None
    extra: Optional[ExtraField] = None


class MachineInstance(BaseModel):
    machine_id: str
    machine_name: str
    location: str
    private_ips: List[str]
    public_ips: Optional[List[str]] = None
    volumes: List[str] = []
    labels: Optional[Dict[str, Any]] = None
    extra: Optional[ExtraField] = None


class BlockInstance(BaseModel):
    id: str
    name: str
    size: Union[int, str]
    location: str
    mount: Optional[str] = None
    snapshot: Optional[str] = None
    image: Optional[str] = None
    permissions: Optional[List[str]] = None
    description: Optional[str] = None
    kind: Optional[str] = None
    extra: Optional[ExtraField] = None


class ExecutionMachine(BaseModel):
    execid: str
    machine_name: str
    machine: MachineRequest
    provider: str
    cluster: str = "default"
    qnames: List[str] = []
    agent_homedir: str = defaults.AGENT_HOMEDIR
    worker_procs: int = 1
    ssh_key: Optional[SSHKey] = None
    docker_uid: str = defaults.DOCKER_UID
    docker_gid: str = defaults.DOCKER_GID
    docker_image: str = defaults.AGENT_DOCKER_IMG
    docker_version: str = "latest"


class ExecMachineResult(BaseModel):
    execid: str
    private_ips: List[str]
    public_ips: Optional[List[str]] = None
    node: MachineInstance



import json
import logging
from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Union

from labfunctions import defaults, errors, secrets, types
from labfunctions.utils import parse_var_line

from .base import BaseClient
from .utils import get_private_key, store_credentials_disk, store_private_key


class HistoryClient(BaseClient):
    """Is to be used as cli client because it has side effects on local disk"""

    def history_register(self, exec_result: types.ExecutionResult) -> bool:

        rsp = self._http.post(
            f"/history",
            json=exec_result.dict(),
        )

        if rsp.status_code == 201:
            return True
        return False

    def history_get_last(
        self, wfid: Optional[str] = None, last=1
    ) -> List[types.HistoryResult]:
        query = f"/history/{self.projectid}?lt={last}"
        if wfid:
            query = f"/history/{self.projectid}/{wfid}?lt={last}"
        rsp = self._http.get(query)
        rows = []
        for r in rsp.json()["rows"]:
            h = types.HistoryResult(**r)
            h.result = types.ExecutionResult(**r["result"])
            rows.append(h)

        return rows

    def history_detail(self, execid: str) -> Union[types.HistoryResult, None]:
        query = f"/history/{self.projectid}/detail/{execid}"
        rsp = self._http.get(query)
        if rsp.status_code == 200:
            return types.HistoryResult(**rsp.json())
        return None

    def history_get_output(self, uri) -> Generator[bytes, None, None]:
        """uri if ok:
        uri = f"{row.result.output_dir}/{row.result.output_name}"
        else:
        uri = f"{row.result.error_dir}/{row.result.output_name}"
        """
        url = f"/history/{self.projectid}/_get_output?file={uri}"
        with self._http.stream("GET", url) as r:
            for data in r.iter_bytes():
                yield data

    def history_nb_output(self, exec_result: types.ExecutionResult) -> bool:
        """Upload the notebook from the execution result
        TODO: zip or compress notebook before upload.
        :return: True if ok, False if something fails.
        """
        form_data = dict(output_name=exec_result.output_name)

        file_dir = f"{exec_result.output_dir}/{exec_result.output_name}"
        _addr = f"/history/{exec_result.projectid}/_output_ok"
        if exec_result.error:
            _addr = f"/history/{exec_result.projectid}/_output_fail"
            file_dir = f"{exec_result.error_dir}/{exec_result.output_name}"

        files = {"file": open(file_dir)}
        rsp = self._http.post(
            _addr,
            files=files,
            data=form_data,
        )
        if rsp.status_code == 201:
            return True
        return False



from dataclasses import dataclass
from typing import List, Optional

from pydantic import BaseModel


class ProjectData(BaseModel):
    name: str
    projectid: str
    # username: Optional[str] = None
    owner: Optional[str] = None
    agent: Optional[str] = None  # to deprecate
    users: Optional[List[str]] = None
    description: Optional[str] = None
    repository: Optional[str] = None

    class Config:
        orm_mode = True


@dataclass
class ProjectReq:
    name: str
    private_key: str
    projectid: Optional[str] = None
    description: Optional[str] = None
    repository: Optional[str] = None


class ProjectBuildReq(BaseModel):
    """
    In the future remote path could be offered like S3, GoogleStore...
    and so on, right now is managed by the server.
    """

    name: str
    server_handle: bool = True
    remote_path: Optional[str] = None


class ProjectBuildResp(BaseModel):
    msg: str
    execid: str


class ProjectCreated(BaseModel):
    pd: ProjectData
    private_key: str



class CommandExecutionException(Exception):
    def __init__(self, message):
        super().__init__(message)


class WorkflowDisabled(Exception):
    def __init__(self, projectid, wfid):
        _msg = f"projectid: {projectid} and wfid: {wfid} disabled"
        super().__init__(_msg)


class WorkflowNotFound(Exception):
    def __init__(self, projectid, wfid):
        _msg = f"projectid: {projectid} and wfid: {wfid} not found"
        super().__init__(_msg)


class WorkflowRegisterError(Exception):
    def __init__(self, project, alias, e):
        if "_nb_workflow__project_alias" in str(e):
            _msg = f"Alias {alias} for {project} already exist"
        else:
            _msg = f"Error registering {alias} for project {project}"
        super().__init__(_msg)


class WorkflowRegisterClientError(Exception):
    def __init__(self, project, wfid):
        _msg = f"Registration error for workflow {wfid} in {project}"
        super().__init__(_msg)


class ProjectNotFound(Exception):
    def __init__(self, message):
        super().__init__(message)


class PrivateKeyNotFound(Exception):
    def __init__(self, projectid):
        super().__init__(f"Private Key not found for project {projectid}")


class AuthValidationFailed(Exception):
    pass



import json

from rich.console import Console
from rich.progress import Progress, SpinnerColumn

from labfunctions.client.diskclient import DiskClient
from labfunctions.utils import format_bytes

console = Console()
progress = Progress(
    SpinnerColumn(),
    "[progress.description]{task.description}",
)


def watcher(c: DiskClient, execid, stats=False):
    keep = True
    last = None

    console = Console()
    console.print(f"[bold magenta]Watching for execid: {execid}[/]")
    events = 0

    for evt in c.events_listen(execid, last=last):
        if evt.event == "stats" and stats:
            _stats = json.loads(evt.data)
            _mem = _stats["mem"]["mem_usage"]
            mem = format_bytes(_mem)
            msg = f"[orange]Memory used {mem}[/]"
            console.print(msg)
        elif evt.event == "result":
            keep = False
            console.print(f"[bold green]Finished execid: [/] [bold magenta]{execid}[/]")
        elif evt.event != "stats":
            msg = evt.data
            console.print(msg)
        last = evt.id
        events += 1



# pylint: disable=unused-argument
import glob
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import List, Optional

import aiofiles
from redis import Redis
from sanic import Blueprint, Sanic, exceptions
from sanic.response import json
from sanic_ext import openapi
from sanic_jwt import protected

from labfunctions.conf.server_settings import settings
from labfunctions.core.core import nb_job_executor
from labfunctions.core.entities import (
    ExecutionResult,
    HistoryRequest,
    NBTask,
    ProjectData,
)
from labfunctions.core.managers import history, projects
from labfunctions.core.registers import register_history_db
from labfunctions.core.scheduler_deprecated import (
    QueueExecutor,
    SchedulerExecutor,
    scheduler_dispatcher,
)
from labfunctions.defaults import API_VERSION
from labfunctions.utils import (
    get_query_param,
    parse_page_limit,
    run_async,
    secure_filename,
)

rqjobs_bp = Blueprint("rqjobs", url_prefix="rqjobs", version=API_VERSION)


def _get_scheduler(qname="default") -> SchedulerExecutor:

    current_app = Sanic.get_app("labfunctions")
    r = current_app.ctx.rq_redis
    return SchedulerExecutor(r, qname=qname)


def _get_q_executor(qname="default") -> QueueExecutor:
    current_app = Sanic.get_app("labfunctions")
    r = current_app.ctx.rq_redis

    return QueueExecutor(r, qname=qname)


def list_workflows():
    notebooks = []
    files = glob(f"{settings.BASE_PATH}/{settings.NB_WORKFLOWS}*")
    for x in files:
        if ".ipynb" or ".py" in x:
            notebooks.append(x.split("/")[-1].split(".")[0])
    return notebooks


@dataclass
class JobResponse:
    wfid: str
    workflow: NBTask


@dataclass
class JobDetail:
    wfid: str
    func_name: str
    # workflow: NBTask
    created_at: str


@rqjobs_bp.get("/")
@openapi.response(200, List[JobDetail], "Task Scheduled")
@protected()
def list_scheduled_redis(request):
    """
    List the jobs scheduled in the scheduler
    """
    # pylint: disable=unused-argument

    scheduler = _get_scheduler()
    jobs = scheduler.list_jobs()

    return json(jobs, 200)


@rqjobs_bp.delete("/_cancel/<wfid>")
@openapi.parameter("wfid", str, "path")
@protected()
async def cancel_job(request, wfid):
    """delete a scheduler job from redis"""
    # pylint: disable=unused-argument

    scheduler = _get_scheduler()
    await run_async(scheduler.cancel_job, wfid)
    return json(dict(msg="done"), 200)


@rqjobs_bp.delete("/_cancel_all")
@protected()
async def schedule_cancel_all(request):
    """Cancel all the jobs in the queue"""
    # pylint: disable=unused-argument

    scheduler = _get_scheduler()

    await run_async(scheduler.cancel_all)

    return json(dict(msg="done"), 200)


@rqjobs_bp.get("/<wfid>")
@openapi.parameter("wfid", str, "path")
@protected()
def get_job_result(request, wfid):
    """Get job result from the queue"""
    Q = _get_q_executor()
    job = Q.fetch_job(wfid)
    result = job.result
    if result:
        result = asdict(result)

    return json(
        dict(
            wfid=job.id,
            status=job.get_status(),
            result=result,
            position=job.get_position(),
        )
    )


@rqjobs_bp.get("/failed")
@protected()
def get_failed_jobs(request):
    """Get jobs failed in RQ"""
    Q = _get_q_executor()

    jobs = Q.get_jobs_ids("failed")

    return json(dict(rows=jobs, total=len(jobs)))


@rqjobs_bp.delete("/failed")
@openapi.parameter("remove", bool, "query")
@protected()
def delete_failed_jobs(request):
    """Remove failed jobs from the queue"""

    to_remove = get_query_param(request, "remove", False)
    Q = _get_q_executor()

    jobs = Q.remove_jobs("failed", to_remove)

    return json(dict(rows=jobs, total=len(jobs)))


@rqjobs_bp.get("/running")
@protected()
def get_running_jobs(request):
    """Get jobs Running"""
    Q = _get_q_executor()

    jobs = Q.get_jobs_ids("started")

    return json(dict(rows=jobs, total=len(jobs)))



# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())


__version__ = "0.7.0-alpha.0"



import io
import os
from datetime import datetime, timedelta
from typing import Any, AsyncGenerator, Dict, Generator, Union

from google.cloud.storage import Client
from smart_open import open

from labfunctions.utils import run_async

from .kvspec import AsyncKVSpec, GenericKVSpec


class KVGS(GenericKVSpec):
    """https://googleapis.dev/python/storage/latest/client.html"""

    def __init__(self, bucket: str, client_opts: Dict[str, Any] = {}):
        self._opts = client_opts
        self._bucket = bucket
        service_account_path = os.environ["GOOGLE_APPLICATION_CREDENTIALS"]
        self.client = Client.from_service_account_json(service_account_path)
        self.bucket = self.client.get_bucket(bucket)
        self.params = {"client": self.client}

    @property
    def uri(self):
        return f"gs://{self._bucket}"

    def put(self, key: str, bdata: bytes):
        # obj = io.BytesIO(bdata)
        blob = self.bucket.blob(key)
        blob.upload_from_string(bdata, content_type="application/octet-stream")

    def _writer(self, key: str, generator: Generator[bytes, None, None]):
        with open(f"{self.uri}/{key}", "wb", transport_params=self.params) as f:
            data = True
            while data:
                try:
                    _data = generator.__next__()
                    f.write(_data)
                except StopIteration:
                    data = False

    def put_stream(self, key: str, generator: Generator[bytes, None, None]) -> bool:
        rsp = True
        try:
            self._writer(key, generator)
        except Exception:
            rsp = False
        return rsp

    def get(self, key: str) -> Union[bytes, None]:
        blob = self.bucket.blob(key)
        obj = None
        try:
            obj = blob.download_as_bytes()
        except Exception:
            pass
        return obj

    def get_stream(self, key: str) -> Generator[bytes, None, None]:
        uri = f"{self.uri}/{key}"
        for chunk in open(uri, "rb", transport_params=self.params):
            yield chunk


class AsyncKVGS(AsyncKVSpec):
    """A hacky solution because thereisn't trustworthy async lib"""

    def __init__(self, bucket: str, client_opts: Dict[str, Any] = {}):
        self._opts = client_opts
        self._bucket = bucket
        self.client = KVGS(bucket, client_opts)

    async def put(self, key: str, bdata: bytes):
        await run_async(self.client.put, key, bdata)

    async def put_stream(
        self, key: str, generator: Generator[bytes, None, None]
    ) -> bool:
        rsp = await run_async(self.client.put_stream, key, generator)
        return rsp

    async def get(self, key: str) -> Union[bytes, str, None]:
        rsp = await run_async(self.client.get, key)
        return rsp

    async def get_stream(self, key: str) -> AsyncGenerator[bytes, None]:
        yield await run_async(self.client.get_stream, key)



from typing import Dict, Union

from labfunctions.types import ServerSettings
from labfunctions.types.cluster import ClusterSpec
from labfunctions.types.machine import BlockStorage, MachineOrm, SSHKey
from labfunctions.utils import open_yaml


def get_local_machine(name, data) -> MachineOrm:
    return MachineOrm(**data["machines"][name])


def get_local_volume(name, data) -> BlockStorage:
    return BlockStorage(**data["volumes"][name])


def ssh_from_settings(settings: ServerSettings) -> Union[SSHKey, None]:

    if settings.CLUSTER_SSH_PUBLIC_KEY:
        ssh = SSHKey(
            user=settings.CLUSTER_SSH_KEY_USER,
            public_path=settings.CLUSTER_SSH_PUBLIC_KEY,
        )

        ssh.private_path = ssh.public_path.split(".pub")[0]
        return ssh
    return None



from typing import List

import asyncssh
from asyncssh.process import SSHCompletedProcess


async def run_cmd(
    addr: str, cmd: str, keys: List[str], username="op", known_hosts=None, check=False
) -> SSHCompletedProcess:
    """
    A opinated wrapper around asyncssh
    :param addr: ip address to connect
    :param cmd: a simple string with the command to run
    :param keys: a List of keys names to be used without extension
    :param username: user to use for ssh connection
    :param known_hosts: validate server host, if None this check will ignored
    :param check: if true an a remote error happend it will raise an exceptions
    """

    async with asyncssh.connect(
        addr, username=username, client_keys=keys, known_hosts=known_hosts
    ) as conn:
        result = await conn.run(cmd, check=check)
    return result


async def scp_from_local(
    remote_addr: str,
    remote_dir: str,
    local_file: str,
    keys: List[str],
    username="op",
    known_hosts=None,
):
    """
    Copy a local file to remote server
    :param remote_addr: ip address to connect
    :param remote_dir: dir where files should be copied
    :param local_file: local file to be copied
    :param keys: a List of keys names to be used without extension
    :param username: user to use for ssh connection
    :param known_hosts: validate server host, if None this check will ignored
    """

    async with asyncssh.connect(
        remote_addr, username=username, client_keys=keys, known_hosts=known_hosts
    ) as conn:
        await asyncssh.scp(local_file, (conn, remote_dir))



from contextvars import ContextVar
from importlib import import_module
from typing import List

import aioredis
import redis
from sanic import Sanic
from sanic.response import json
from sanic_ext import Extend

from labfunctions import defaults
from labfunctions.db.nosync import AsyncSQL
from labfunctions.events import EventManager
from labfunctions.io.kvspec import AsyncKVSpec
from labfunctions.security import auth_from_settings, sanic_init_auth
from labfunctions.security.redis_tokens import RedisTokenStore
from labfunctions.types import ServerSettings
from labfunctions.utils import get_class, get_version

version = get_version("__version__.py")


def init_blueprints(app, blueprints_allowed, package_dir="labfunctions.web"):
    """
    It will import bluprints from modules that ends with "_bp" and belongs
    to the package declared in `changeme.defaults.SANIC_BLUPRINTS_DIR`
    by default it will be `changeme.services`
    """
    blueprints = set()
    mod = app.__module__
    for mod_name in blueprints_allowed:
        modules = import_module(f"{package_dir}.{mod_name}_bp", mod)
        for el in dir(modules):
            if el.endswith("_bp"):
                bp = getattr(modules, el)
                blueprints.add(bp)

    for bp in blueprints:
        print("Adding blueprint: ", bp.name)
        app.blueprint(bp)


def create_db_instance(url) -> AsyncSQL:
    return AsyncSQL(url)


def create_web_redis(url):
    return aioredis.from_url(url, decode_responses=True)


def create_rq_redis(url):
    return redis.from_url(url)


def create_projects_store(store_class, store_bucket) -> AsyncKVSpec:
    Class = get_class(store_class)
    return Class(store_bucket)


def create_app(
    settings: ServerSettings,
    list_bp: List[str],
    app_name=defaults.SANIC_APP_NAME,
    db_func=create_db_instance,
    web_redis_func=create_web_redis,
    rq_func=create_rq_redis,
    projects_store_func=create_projects_store,
    with_auth=True,
    with_auth_bp=True,
) -> Sanic:
    """Factory pattern like flask"""

    app = Sanic(app_name)

    app.config.CORS_ORIGINS = "*"

    Extend(app)
    app.ext.openapi.add_security_scheme(
        "token",
        "http",
        scheme="bearer",
        bearer_format="JWT",
    )
    # app.ext.openapi.secured()
    app.ext.openapi.secured("token")

    _base_model_session_ctx = ContextVar("session")

    init_blueprints(app, list_bp)

    web_redis = web_redis_func(settings.WEB_REDIS)
    if with_auth:
        _store = RedisTokenStore(web_redis)
        auth = auth_from_settings(settings.SECURITY, _store)
        sanic_init_auth(app, auth, settings.SECURITY)
    if with_auth and with_auth_bp:
        init_blueprints(app, ["auth"], "labfunctions.security")

    @app.listener("before_server_start")
    async def startserver(current_app, loop):
        """This function runs one time per worker"""
        _db = db_func(settings.ASQL)
        _base_model_session_ctx = ContextVar("session")

        current_app.ctx.kv_store = projects_store_func(
            settings.PROJECTS_STORE_CLASS_ASYNC, settings.PROJECTS_STORE_BUCKET
        )
        current_app.ctx.web_redis = web_redis.client()
        current_app.ctx.rq_redis = rq_func(settings.RQ_REDIS)
        current_app.ctx.db = _db
        await current_app.ctx.db.init()

    @app.middleware("request")
    async def inject_session(request):
        current_app = Sanic.get_app(defaults.SANIC_APP_NAME)

        request.ctx.session = current_app.ctx.db.sessionmaker()
        request.ctx.session_ctx_token = _base_model_session_ctx.set(request.ctx.session)
        request.ctx.web_redis = current_app.ctx.web_redis
        request.ctx.events = EventManager(current_app.ctx.web_redis)

        request.ctx.dbconn = current_app.ctx.db.engine

    @app.middleware("response")
    async def close_session(request, response):
        if hasattr(request.ctx, "session_ctx_token"):
            _base_model_session_ctx.reset(request.ctx.session_ctx_token)
            await request.ctx.session.close()

    @app.listener("after_server_stop")
    async def shutdown(current_app, loop):
        await current_app.ctx.db.engine.dispose()
        # await current_app.ctx.redis.close()

    @app.get("/status")
    async def status_handler(request):
        return json(dict(msg="We are ok", version=version))

    return app



from collections import OrderedDict
from typing import Dict, List, Optional, Union

import yaml

from labfunctions import runtimes
from labfunctions.types import NBTask, ProjectData, WorkflowDataWeb
from labfunctions.types.client import WorkflowsFile

# from labfunctions.types.runtimes import RuntimeData
from labfunctions.utils import Singleton, open_yaml, write_yaml


class WFDumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super(WFDumper, self).increase_indent(flow, False)


class WorkflowsState:
    """This manage workflows state, is NOT Thread safe"""

    def __init__(
        self,
        project: Optional[ProjectData] = None,
        workflows: Optional[Dict[str, WorkflowDataWeb]] = None,
        # runtimes: Optional[List[RuntimeData]] = None,
        version="0.2.0",
        workflow_file: Optional[str] = None,
    ):
        self._version = version
        self._project = project
        self._workflows = workflows or {}
        self._file = workflow_file
        # self._runtimes = runtimes

    # @property
    # def runtimes(self) -> Union[List[RuntimeData], None]:
    #     return self._runtimes

    @property
    def workflows_file(self) -> Union[str, None]:
        return self._file

    @property
    def projectid(self) -> Union[str, None]:
        if self.project:
            return self.project.projectid
        return None

    @projectid.setter
    def projectid(self, projectid):
        if not self.project:
            raise AttributeError("No project property")
        self.project.projectid = projectid

    @property
    def project_name(self) -> Union[str, None]:
        if self.project:
            return self.project.name
        return None

    @staticmethod
    def listworkflows2dict(
        workflows: List[WorkflowDataWeb],
    ) -> Dict[str, WorkflowDataWeb]:
        _workflows = {w.alias: w for w in workflows}
        return _workflows

    def add_workflow(self, wf: WorkflowDataWeb):
        self._workflows.update({wf.alias: wf})

    def delete_workflow(self, alias):
        del self._workflows[alias]

    def find_by_id(self, wfid) -> Union[WorkflowDataWeb, None]:
        for alias in self._workflows:
            if self._workflows[alias].wfid == wfid:
                return self._workflows[alias]
        return None

    def update_project(self, pd: ProjectData):
        self._project = pd

    @property
    def file(self) -> WorkflowsFile:
        return WorkflowsFile(
            version=self._version,
            project=self._project,
            workflows=self._workflows,
            # runtime=self._runtime,
        )

    @property
    def workflows(self) -> Dict[str, WorkflowDataWeb]:
        return self._workflows

    @property
    def project(self) -> ProjectData:
        return self._project

    def snapshot(self) -> WorkflowsFile:
        """
        It makes a deep copy of the workfile"
        """
        wf = self.file.copy(deep=True)

        # self.snapshots.append(wf)
        return wf

    @staticmethod
    def read(filepath="workflows.yaml") -> WorkflowsFile:
        # data_dict = open_toml(filepath)
        data_dict = open_yaml(filepath)

        wf = WorkflowsFile(**data_dict, workflows_file=filepath)
        return wf

    def write(self, fp="workflows.yaml"):
        """
        Writes the state to workflows.yaml (by default)
        Also perfoms some serializations
        """
        wf = self.snapshot()
        wf_dict = wf.dict(exclude_none=True)
        _dict = OrderedDict()
        _dict["version"] = wf.version
        _dict["project"] = wf.project.dict(exclude_none=True)
        # if wf_dict.get("runtime"):
        #    _dict["runtime"] = wf.runtime.dict(exclude_none=True)
        if wf_dict.get("workflows"):
            _dict["workflows"] = {
                k: v.dict(exclude_none=True, exclude_unset=True)
                for k, v in wf.workflows.items()
            }
        write_yaml(
            fp, dict(_dict), Dumper=WFDumper, default_flow_style=False, sort_keys=False
        )


def from_file(fpath="workflows.yaml", runtimes_path="runtimes.yaml") -> WorkflowsState:
    # runtimes_data = open_yaml(runtimes_path)
    # runtimes = [RuntimeData ]
    wf = WorkflowsState.read(fpath)
    return WorkflowsState(
        project=wf.project,
        version=wf.version,
        workflows=wf.workflows,
        # runtime=wf.runtime,
    )



import httpx

SLACK_API = "https://slack.com/api/"
DISCORD_API = "https://discord.com/api/webhooks/"
EMOJI_ERROR = "🤬"
EMOJI_OK = "👌"


class SlackCient:
    def __init__(self, tkn, addr=SLACK_API):
        self._headers = {"Authorization": f"Bearer {tkn}"}
        self._addr = addr

    def list_channels(self):
        r = httpx.get(f"{self._addr}/conversations.list", headers=self._headers)
        return r.json()

    def send(self, channel, text):
        r = httpx.post(
            f"{self._addr}/chat.postMessage",
            headers=self._headers,
            data=dict(channel=channel, text=text),
        )
        return r.json()


class DiscordClient:
    def __init__(self, addr=DISCORD_API):
        self._addr = addr

    def send(self, channel, text, username="NB Workflows"):
        fullurl = f"{self._addr}{channel}"
        r = httpx.post(fullurl, json={"content": text, "username": username})
        return r.text



from sanic import Blueprint, Request
from sanic.response import empty, json, stream
from sanic_ext import openapi

from labfunctions.security import get_auth, get_authenticate, protected
from labfunctions.types.security import JWTResponse, UserLogin

from .errors import AuthValidationFailed, MissingAuthorizationHeader, WebAuthFailed

auth_bp = Blueprint("auth_api", url_prefix="auth", version="v1")


@auth_bp.post("/login")
@openapi.response(200, {"application/json": JWTResponse})
@openapi.response(403, dict(msg=str), "Not Found")
@openapi.body(UserLogin)
async def login_handler(request):
    auth = get_auth(request)
    authenticate = get_authenticate(request)
    try:
        user = await authenticate(request)
    except AuthValidationFailed:
        raise WebAuthFailed()
    encoded = auth.encode({"usr": user.username, "scopes": user.scopes.split(",")})

    rtkn = await auth.store_refresh_token(user.username)
    return json(JWTResponse(access_token=encoded, refresh_token=rtkn).dict(), 200)


@auth_bp.get("/verify")
@openapi.response(200, {"application/json": JWTResponse})
@protected()
async def verify_handler(request):
    return json(request.ctx.token_data, 200)


@auth_bp.post("/refresh_token")
@openapi.response(200, {"application/json": JWTResponse})
@openapi.body(JWTResponse)
async def refresh_handler(request):
    if not request.app.config.AUTH_ALLOW_REFRESH:
        return json(dict(msg="Not found"), 404)

    at = request.token
    rftkn = request.json.get("refresh_token")
    if not rftkn:
        raise WebAuthFailed()

    old_token = JWTResponse(access_token=at, refresh_token=rftkn)
    # redis = request.ctx.web_redis
    auth = get_auth(request)
    try:
        jwt_res = await auth.refresh_token(
            old_token.access_token, old_token.refresh_token
        )
        return json(jwt_res.dict(), 200)
    except AuthValidationFailed as e:
        print(e)
        raise WebAuthFailed()



from pathlib import Path
from typing import Any, Dict, Optional, Union

from labfunctions import defaults
from labfunctions.conf.jtemplates import get_package_dir, render_to_file
from labfunctions.types.runtimes import DockerSpec, RuntimeData, RuntimeSpec
from labfunctions.utils import execute_cmd, open_yaml


def generate_dockerfile(dst_root: Path, runtime: RuntimeSpec):
    render_to_file(
        runtime.container.base_template,
        str((dst_root / f"Dockerfile.{runtime.name}").resolve()),
        data=runtime.container.dict(),
    )


def get_runtimes_specs(from_file="runtimes.yaml") -> Dict[str, RuntimeSpec]:
    data = open_yaml(from_file)
    runtimes = {k: RuntimeSpec(name=k, **v) for k, v in data["runtimes"].items()}
    return runtimes


def get_spec_from_file(name: str, from_file="runtimes.yaml") -> RuntimeSpec:
    specs = get_runtimes_specs(from_file)
    return specs[name]


def git_short_head_id():
    return execute_cmd("git rev-parse --short HEAD")


def git_last_tag():
    return execute_cmd("git describe --tags")



# pylint: disable=unused-argument
import json as std_json
import pathlib
from dataclasses import asdict

import httpx
from sanic import Blueprint
from sanic.response import json
from sanic_ext import openapi

from labfunctions import defaults
from labfunctions.conf.server_settings import settings
from labfunctions.defaults import API_VERSION
from labfunctions.managers import history_mg
from labfunctions.managers.users_mg import inject_user
from labfunctions.security.web import protected
from labfunctions.types import ExecutionResult, HistoryRequest, NBTask
from labfunctions.utils import today_string
from labfunctions.web.utils import get_kvstore, get_query_param2

history_bp = Blueprint("history", url_prefix="history", version=API_VERSION)

# async def validate_project(request):
#     request.ctx.user = await extract_user_from_request(request)


@history_bp.post("/")
@openapi.body({"application/json": ExecutionResult})
@openapi.response(201, "Created")
@protected()
async def history_create(request):
    """Register a jobexecution"""
    # pylint: disable=unused-argument
    dict_ = request.json
    exec_result = ExecutionResult(**dict_)

    session = request.ctx.session
    async with session.begin():
        hm = await history_mg.create(session, exec_result)

    return json(dict(msg="created"), 201)


@history_bp.get("/<projectid>")
@openapi.parameter("projectid", str, "path")
@openapi.response(200, "Found")
@openapi.response(404, dict(msg=str), "Not Found")
@openapi.parameter("lt", int, "lt")
@protected()
async def history_get_all(request, projectid):
    """Get the status of the last job executed"""
    # pylint: disable=unused-argument
    lt = get_query_param2(request, "lt", 1)
    session = request.ctx.session
    async with session.begin():
        h = await history_mg.get_last(session, projectid, limit=lt)
        if h.rows:
            return json(h.dict(), 200)

    return json(dict(msg="not found"), 404)


@history_bp.get("/<projectid>/detail/<execid>")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("execid", str, "path")
@openapi.response(200, "Found")
@openapi.response(404, dict(msg=str), "Not Found")
@protected()
async def history_detail_job(request, projectid: str, execid: str):
    """Get the status of the last job executed"""
    # pylint: disable=unused-argument
    session = request.ctx.session
    async with session.begin():
        h = await history_mg.get_one(session, execid)
        if h:
            return json(h.dict(), 200)

        return json(dict(msg="not found"), 404)


@history_bp.get("/<projectid>/<wfid>")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("wfid", str, "path")
@openapi.response(200, "Found")
@openapi.response(404, dict(msg=str), "Not Found")
@openapi.parameter("lt", int, "lt")
@protected()
async def history_last_job(request, wfid, projectid):
    """Get the status of the last job executed"""
    # pylint: disable=unused-argument
    lt = get_query_param2(request, "lt", 1)
    session = request.ctx.session
    async with session.begin():
        h = await history_mg.get_last(session, projectid, wfid, limit=lt)
        if h.rows:
            return json(h.dict(), 200)

        return json(dict(msg="not found"), 404)


@history_bp.post("/<projectid>/_output_ok")
@openapi.parameter("projectid", str, "path")
@protected()
async def history_output_ok(request, projectid):
    """
    Upload a workflow project
    """
    # pylint: disable=unused-argument
    kv_store = get_kvstore(request)

    today = today_string(format_="day")
    root = pathlib.Path(projectid)
    output_dir = root / defaults.NB_OUTPUTS / "ok" / today

    file_body = request.files["file"][0].body
    output_name = request.form["output_name"][0]

    fp = str(output_dir / output_name)
    # await fsrv.put(fp, file_body)
    await kv_store.put(fp, file_body)

    return json(dict(msg="OK"), 201)


@history_bp.post("/<projectid>/_output_fail")
@openapi.parameter("projectid", str, "path")
@protected()
async def history_output_fail(request, projectid):
    """
    Upload a workflow project
    """
    # pylint: disable=unused-argument

    kv_store = get_kvstore(request)

    today = today_string(format_="day")
    root = pathlib.Path(projectid)
    output_dir = root / defaults.NB_OUTPUTS / "errors" / today

    file_body = request.files["file"][0].body
    output_name = request.form["output_name"][0]

    fp = str(output_dir / output_name)
    await kv_store.put(fp, file_body)

    return json(dict(msg="OK"), 201)


@history_bp.get("/<projectid>/_get_output")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("file", str, "query")
@protected()
async def history_get_output(request, projectid):
    """
    Upload a workflow project
    """
    # pylint: disable=unused-argument
    uri = request.args.get("file")
    key = f"{projectid}/{uri}"
    response = await request.respond(content_type="application/octet-stream")
    kv_store = get_kvstore(request)
    async for chunk in kv_store.get_stream(key):
        await response.send(chunk)
    await response.eof()



import asyncio
import json as jsondef

import aioredis
import async_timeout
from sanic import Blueprint
from sanic.response import empty, json, stream
from sanic_ext import openapi

from labfunctions.defaults import API_VERSION
from labfunctions.events import EventManager
from labfunctions.security.web import protected
from labfunctions.types.events import EventSSE
from labfunctions.utils import get_query_param, secure_filename

events_bp = Blueprint("events", url_prefix="events", version=API_VERSION)


def get_event_manager(request) -> EventManager:
    return request.ctx.events


@events_bp.get("/<projectid>/<execid>/_listen")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("execid", str, "path")
@openapi.parameter("last", str, "query")
@openapi.response(200, "Found")
@openapi.response(404, dict(msg=str), "Not Found")
@protected()
async def event_listen(request, projectid, execid):
    # redis = get_web_redis(request)
    em = get_event_manager(request)
    last = request.args.get("last", "$")

    response = await request.respond(
        content_type="text/event-stream", headers={"Cache-Control": "no-store"}
    )

    channel = EventManager.generate_channel(projectid, execid)
    keep = True
    while keep:
        rsp = await em.read(channel, last)
        if rsp:
            for r in rsp:
                if r.event == "control" and r.data == "exit":
                    keep = False
                msg = em.format_sse(r)
                await response.send(msg)
            last = r.id
        else:
            keep = False

    await response.eof()


@events_bp.post("/<projectid>/<execid>/_publish")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("execid", str, "path")
@openapi.response(204)
@openapi.response(404, dict(msg=str), "Not Found")
@protected()
async def event_publish(request, projectid, execid):
    evt = EventSSE(**request.json)

    em = get_event_manager(request)
    channel = EventManager.generate_channel(projectid, execid)
    await em.publish(channel, evt)

    return empty()



import os
import sys

AGENT_TOKEN_EXP = (60 * 60) * 12
# Services
SQL = os.getenv("NB_SQL", "sqlite:///db.sqlite")
ASQL = os.getenv("NB_ASQL", "sqlite+aiosqlite:///db.sqlite")
WORKFLOW_SERVICE = os.getenv("NB_WORKFLOW_SERVICE", "http://localhost:8000")

# General Folders for the server
BASE_PATH = os.getenv("NB_BASEPATH", os.getcwd())
AGENT_DATA_FOLDER = ".worker_data/"
SECURITY = {
    "JWT_ALG": "ES512",
    "JWT_EXP": 30,
    "AUTH_CLASS": "labfunctions.security.authentication.Auth",
    "AUTH_FUNCTION": "labfunctions.managers.users_mg.authenticate",
}

# Logs
LOGLEVEL = "INFO"
LOGCONFIG = dict(  # no cov
    version=1,
    disable_existing_loggers=False,
    loggers={
        "nbwork.server": {"level": LOGLEVEL, "handlers": ["console"]},
        "nbwork.error": {
            "level": LOGLEVEL,
            "handlers": ["error_console"],
            "propagate": True,
            "qualname": "nbwork.error",
        },
    },
    handlers={
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "generic",
            "stream": sys.stdout,
        },
        "error_console": {
            "class": "logging.StreamHandler",
            "formatter": "generic",
            "stream": sys.stderr,
        },
    },
    formatters={
        "generic": {
            "format": "%(asctime)s [%(process)d] [%(levelname)s] %(message)s",
            "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
            "class": "logging.Formatter",
        },
    },
)



import json
import os
import sys

import click
from rich import print_json

from labfunctions import client, defaults
from labfunctions.client.diskclient import DiskClient
from labfunctions.conf import load_client
from labfunctions.context import create_dummy_ctx
from labfunctions.executors import jupyter_exec
from labfunctions.executors.docker_exec import docker_exec
from labfunctions.executors.local_exec import local_exec_env
from labfunctions.hashes import generate_random
from labfunctions.types import NBTask

from .utils import console, watcher


def _parse_params_args(params):
    params_dict = {}
    for p in params:
        k, v = p.split("=")
        params_dict.update({k: v})
    return params_dict


@click.group(name="exec")
def executorscli():
    """
    Execute workflows or notebooks, locally or remote
    """


@executorscli.command()
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--url-service",
    "-u",
    default=load_client().WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option("--wfid", "-W", default=None, help="Workflow ID to execute")
@click.option("--dev", "-d", default=False, is_flag=True, help="Execute locally")
def local(from_file, url_service, dev, wfid):
    """Used by the agent to run workloads or for development purposes"""
    rsp = None

    if os.environ.get(defaults.EXECUTIONTASK_VAR):
        console.print(f"=> Starting work")
        rsp = local_exec_env()

    elif wfid:
        c = client.from_file(from_file, url_service=url_service)
        exec_task = c.build_context(wfid)
        os.environ[defaults.EXECUTIONTASK_VAR] = exec_task.json()
        rsp = local_exec_env()
    else:
        console.print(
            f"[red]Neither [bold magenta]wfid[/bold magenta] param or "
            f"[bold magenta]{defaults.EXECUTIONTASK_VAR}[/bold magenta]"
            " was provided[/]"
        )
        sys.exit(-1)

    if rsp:
        status = "OK"
        if rsp.error:
            console.print(f"[bold red](X) WFID: {rsp.wfid} failed[/]")
            sys.exit(-1)
        console.print(f"=>[bold green] WFID: {rsp.wfid} locally executed[/]")


@executorscli.command()
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--url-service",
    "-u",
    default=load_client().WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option("--wfid", "-W", default=None, help="Workflow ID to execute")
@click.option("--notebook", "-n", default=None, help="Notebook to execute")
@click.argument("action", type=click.Choice(["workflow", "notebook"]))
def docker(from_file, url_service, notebook, wfid, action):
    """It will run the task inside a docker container, it exist only as a tester"""
    if wfid and not notebook:
        c = client.from_file(from_file, url_service=url_service)
        exec_task = c.build_context(wfid)
        os.environ["NB_AGENT_TOKEN"] = c.creds.access_token
        os.environ["NB_AGENT_REFRESH_TOKEN"] = c.creds.refresh_token
        os.environ["NB_WORKFLOW_SERVICE"] = url_service
        result = docker_exec(exec_task)
        print_json(data=result.dict())


# @executorscli.command()
# @click.option(
#    "--watch",
#    "-w",
#    is_flag=True,
#    default=False,
#    help="Get events & logs from the executions",
# )
# @click.option(
#    "--stats",
#    "-s",
#    is_flag=True,
#    default=False,
#    help="Show stats as memory usage from the execution",
# )
# @click.argument("wfid")
# @click.pass_context
# def wf(ctx, wfid, watch, stats):
#    """It will dispatch the workflow task to the server"""
#
#    url_service = ctx.obj["URL"]
#    from_file = ctx.obj["WF_FILE"]
#    c = client.from_file(from_file, url_service=url_service)
#    execid = c.workflows_enqueue(wfid)
#    # if execid:
#    console.print(f"Executed: {execid} on server {url_service}")
#    if watch:
#        watcher(c, execid, stats=stats)
#    else:
#        console.print("[red]Something went wrong[/]")
#
#
@executorscli.command()
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--url-service",
    "-u",
    default=load_client().WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option(
    "--param", "-p", multiple=True, help="Params to be passed to the notebook file"
)
@click.option(
    "--machine", "-m", default="cpu", help="Machine where the notebook should run"
)
@click.option("--runtime", "-r", default=None, help="Runtime to use")
@click.option("--cluster", "-c", default="default", help="Cluster where it should run")
@click.option("--version", "-v", default=None, help="Runtime version to run")
@click.option("--dev", "-d", default=False, is_flag=True, help="Execute locally")
@click.option(
    "--watch",
    "-w",
    is_flag=True,
    default=False,
    help="Get events & logs from the executions",
)
@click.argument("notebook")
def notebook(
    url_service,
    from_file,
    param,
    cluster,
    runtime,
    machine,
    version,
    dev,
    notebook,
    watch,
):
    """On demand execution of a notebook file, with custom parameters"""
    # from labfunctions.executors.development import local_nb_dev_exec

    c = client.from_file(from_file, url_service=url_service)
    params_dict = _parse_params_args(param)

    if not dev:
        rsp = c.notebook_run(
            notebook,
            params_dict,
            cluster=cluster,
            machine=machine,
            runtime=runtime,
            version=version,
        )
        # print_json(rsp.json())
        if watch:
            watcher(c, rsp.execid, stats=False)
        print_json(data=rsp.dict())
    else:
        ctx = create_dummy_ctx(c.projectid, notebook, params_dict)
        os.environ[defaults.EXECUTIONTASK_VAR] = ctx.json()
        result = local_exec_env()
        # rsp = local_nb_dev_exec(task)
        print_json(data=result.dict())


@executorscli.command()
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--url-service",
    "-u",
    default=load_client().WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option(
    "--param", "-p", multiple=True, help="Params to be passed to the notebook file"
)
@click.option(
    "--machine", "-m", default="cpu", help="Machine where the notebook should run"
)
@click.option("--runtime", "-r", default=None, help="Runtime to use")
@click.option("--cluster", "-c", default="default", help="Cluster where it should run")
@click.option("--version", "-v", default=None, help="Runtime version to run")
@click.option("--local", "-L", default=False, is_flag=True, help="Execute locally")
def jupyter(
    url_service,
    from_file,
    param,
    cluster,
    runtime,
    machine,
    version,
    local,
):
    """Jupyter instance on demand"""

    c = client.from_file(from_file, url_service=url_service)
    if local:
        os.environ["NB_LOCAL"] = "yes"
        random_url = generate_random(alphabet=defaults.NANO_URLSAFE_ALPHABET)
        opts = jupyter_exec.JupyterOpts(base_url=f"/{random_url}")
        jupyter_exec.jupyter_exec(opts)



from typing import Any, Dict, List, NewType, Optional, Set, Tuple, Union

from pydantic import BaseModel, BaseSettings

from labfunctions import defaults


class DataFolder(BaseModel):
    projectid: str
    name: str
    provider: str
    location: str
    volume_name: str


class ScaleItems(BaseModel):
    """It scale by items enqueue in Redis queue"""

    qname: str
    items_gt: int
    items_lt: int = -1
    increase_by: int = 1
    decrease_by: int = 1
    name: str = "items"


class ScaleIdle(BaseModel):
    """idle_time in minutes"""

    idle_time_gt: int
    idle_time_lt: Optional[int] = None
    name: str = "idle"


class ScaleTimeframe(BaseModel):
    pass


class ClusterState(BaseModel):
    agents_n: int
    agents: Set[str]
    queue_items: Dict[str, int]
    idle_by_agent: Dict[str, int]
    # machines:


class ClusterPolicy(BaseModel):
    min_nodes: int
    max_nodes: int
    strategies: List[Any] = []
    default_nodes: Optional[int] = None


class ClusterDiff(BaseModel):
    to_delete: List[str]
    to_create: int


class ClusterSpec(BaseModel):
    name: str
    provider: str
    machine: str
    location: str
    qnames: List[str]
    policy: ClusterPolicy
    network: str = "default"


class ClusterFile(BaseModel):
    clusters: Dict[str, ClusterSpec]
    inventory: Optional[str] = None
    default_cluster: Optional[str] = None



from typing import List, Optional

from pydantic import BaseModel

from labfunctions import defaults


class AgentNode(BaseModel):
    """
    A self register entity for each machine created
    """

    ip_address: str
    name: str
    pid: str
    qnames: List[str]
    cluster: str
    workers: List[str]
    birthday: int
    machine_id: Optional[str] = None


class AgentConfig(BaseModel):
    redis_dsn: str
    cluster: str
    qnames: List[str]
    ip_address: str
    machine_id: str
    heartbeat_ttl: int
    heartbeat_check_every: int
    agent_name: Optional[str] = None
    workers_n = 1


class AgentRequest(BaseModel):
    machine_ip: str
    machine_id: str
    private_key_path: str
    cluster: str
    qnames: List[str] = ["default"]
    agent_homedir: str = defaults.AGENT_HOMEDIR
    agent_name: Optional[str] = None
    advertise_addr: Optional[str] = None
    docker_image: str = defaults.AGENT_DOCKER_IMG
    docker_version: str = "latest"
    worker_procs: int = 1



import glob
import json
import os
import shutil
import signal
from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseSettings

from labfunctions.types.machine import (
    BlockInstance,
    BlockStorage,
    ExecutionMachine,
    MachineInstance,
    MachineRequest,
)
from labfunctions.utils import (
    execute_cmd_no_block,
    get_external_ip,
    get_internal_ip,
    mkdir_p,
)

from .base import ProviderSpec


class LCConf(BaseSettings):
    working_dir: str

    class Config:
        env_prefix = "NB_LCL_"


class LocalProvider(ProviderSpec):
    def __init__(self, conf: Optional[LCConf] = None):
        self.conf = conf or LCConf()
        self.working_dir = Path(self.conf.working_dir)

    def create_machine(self, node: MachineRequest) -> MachineInstance:
        internal = get_internal_ip()
        external = get_external_ip()
        fp = self.working_dir / node.name
        mkdir_p(fp)

        res = MachineInstance(
            machine_id=f"/local/{node.location}/{node.name}",
            machine_name=node.name,
            location=node.location,
            private_ips=[internal],
            public_ips=[external],
            labels=node.labels,
        )
        with open(fp / "machine.json", "w") as f:
            f.write(res.json())
        return res

    def list_machines(
        self, location: Optional[str] = None, tags: Optional[List[str]] = None
    ) -> List[MachineInstance]:

        machines = []
        for node in glob.glob(f"{self.working_dir}/*/machine.json"):
            with open(node, "r") as f:
                data = f.read()
                obj = json.loads(data)
                instance = MachineInstance(**obj)
                machines.append(instance)

        return machines

    def destroy_machine(self, node: Union[MachineInstance, str]):
        # rm dir: self.working_dir / ctx.node.name
        node_dir = node
        if isinstance(node, MachineInstance):
            node_dir = node.machine_name

        shutil.rmtree(str(self.working_dir / node_dir))

    def create_volume(self, disk: BlockStorage) -> BlockInstance:
        pass

    def destroy_volume(self, disk: Union[str, BlockStorage]) -> bool:
        pass

    def attach_volume(self, node: MachineInstance, disk: BlockStorage) -> bool:
        pass

    def detach_volume(self, node: MachineInstance, disk: BlockStorage) -> bool:
        pass



from .bundler import bundle_project
from .context import (
    build_upload_uri,
    create_build_ctx,
    local_runtime_data,
    make_docker_name,
)
from .utils import generate_dockerfile, get_spec_from_file



import tempfile
import time
from typing import Any, Dict, List, Optional

from asyncssh.process import SSHCompletedProcess
from tenacity import retry, stop_after_attempt, wait_random

from labfunctions import defaults
from labfunctions.cluster import ssh
from labfunctions.cluster.utils import ssh_from_settings
from labfunctions.conf.jtemplates import render_to_file
from labfunctions.types import ServerSettings
from labfunctions.types.agent import AgentRequest
from labfunctions.utils import execute_cmd_no_block, get_version, run_sync


def _prepare_agent_cmd(
    ip_address: str,
    machine_id: str,
    cluster: str,
    qnames: str,
    workers_n=1,
):
    """
    It will run nb agent command.
    Name is not provided, so the agent will choose their name.
    """

    cmd = f"nb agent -i {ip_address} -C {cluster} -q {qnames} -w {workers_n} -m {machine_id}"
    return cmd


def _prepare_docker_cmd(
    ip_address: str,
    machine_id: str,
    qnames: str,
    cluster: str,
    env_file: str,
    docker_image: str,
    docker_version="latest",
    workers_n=1,
):

    nb_agent_cmd = _prepare_agent_cmd(
        ip_address, machine_id, cluster, qnames, workers_n
    )

    cmd = (
        f"docker run -d -v /var/run/docker.sock:/var/run/docker.sock "
        f"-e NB_SERVER=true  --env-file={env_file} "
        f"{docker_image}:{docker_version} "
        f"{nb_agent_cmd}"
    )
    return cmd


@retry(stop=stop_after_attempt(3), wait=wait_random(min=1, max=3))
def agent(req: AgentRequest, data_settings: Dict[str, Any]) -> SSHCompletedProcess:
    """
    Deploy an agent into a server, it has two steps:
    render and copy a .env.docker file into the remote server
    and pull and start the agent's docker instance
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        env_file = f"{tmpdir}/.env.docker"
        render_to_file(defaults.AGENT_ENV_TPL, env_file, data=data_settings)
        run_sync(
            ssh.scp_from_local,
            remote_addr=req.machine_ip,
            remote_dir=req.agent_homedir,
            local_file=env_file,
            keys=[req.private_key_path],
        )

    agent_env_file = f"{req.agent_homedir}/.env.docker"
    addr = req.advertise_addr or req.machine_ip
    cmd = _prepare_docker_cmd(
        addr,
        machine_id=req.machine_id,
        qnames=",".join(req.qnames),
        cluster=req.cluster,
        env_file=agent_env_file,
        docker_image=req.docker_image,
        docker_version=req.docker_version,
        workers_n=req.worker_procs,
    )
    result = run_sync(ssh.run_cmd, req.machine_ip, cmd, keys=[req.private_key_path])
    return result


def agent_local(req: AgentRequest, data_settings: Dict[str, Any], use_docker=False):

    addr = req.advertise_addr or req.machine_ip
    env_file = f"{req.agent_homedir}/.env.docker"
    render_to_file(defaults.AGENT_ENV_TPL, env_file, data=data_settings)

    if use_docker:
        cmd = _prepare_docker_cmd(
            addr,
            machine_id=req.machine_id,
            qnames=",".join(req.qnames),
            cluster=req.cluster,
            env_file=env_file,
            docker_image=req.docker_image,
            docker_version=req.docker_version,
            workers_n=req.worker_procs,
        )
    else:
        cmd = _prepare_agent_cmd(
            addr,
            machine_id=req.machine_id,
            cluster=req.cluster,
            qnames=",".join(req.qnames),
            workers_n=req.worker_procs,
        )

    result = execute_cmd_no_block(cmd, check=False)
    return {"pid": result.pid}


def agent_from_settings(
    ip,
    machine_id,
    cluster,
    settings: ServerSettings,
    qnames: List[str],
    worker_procs=1,
    docker_version=None,
) -> AgentRequest:
    key = ssh_from_settings(settings)
    version = docker_version or get_version()
    return AgentRequest(
        machine_ip=ip,
        machine_id=machine_id,
        private_key_path=key.private_path,
        cluster=cluster,
        qnames=qnames,
        agent_homedir=settings.AGENT_HOMEDIR,
        docker_version=version,
        worker_procs=worker_procs,
    )



import os
from pathlib import Path

import click

# from labfunctions.io.fileserver import FileFileserver
import httpx
from httpx import ConnectError
from rich import print_json
from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm, Prompt

from labfunctions import client
from labfunctions.client import init_script
from labfunctions.conf import load_client
from labfunctions.utils import mkdir_p

service = os.getenv("NS_WORKFLOW_SERVICE", "http://localhost:8000")
console = Console()


@click.command()
@click.option(
    "--url-service",
    "-u",
    default=load_client().WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
def login(url_service):
    """Login to NB Workflows service"""
    # click.echo(f"\nLogin to NB Workflows services {url_service}\n")
    c = client.diskclient.DiskClient(url_service)
    try:
        c.logincli()
        console.print("[bold green]Successfully logged")
    except ConnectError:
        console.print(
            f"[bold red]Error trying to connect to [magenta]{url_service}[/magenta][/]"
        )


@click.command()
@click.option(
    "--create-dirs",
    "-C",
    is_flag=True,
    default=True,
    help="Create outpus and workflows dir",
)
@click.option(
    "--url-service",
    "-u",
    default=service,
    help="URL of the NB Workflow Service",
)
@click.argument("base_path")
def startproject(url_service, create_dirs, base_path):
    """Start a new project"""

    root = Path(base_path)
    console = Console()
    p = Panel.fit(
        "[bold magenta]:smile_cat: Hello and welcome to "
        " NB Workflows [/bold magenta]",
        border_style="red",
    )
    console.print(p)

    should_create = True
    if init_script.i_should_create(root):
        console.print(
            f"\n Starting project at [bold blue underline]{root.resolve()}[/bold blue underline]\n"
        )

        created = False
        confirm = True
        while not created and confirm:
            name = init_script.ask_project_name()
            state = init_script.init_automatic(base_path, name, url_service)
            create = Confirm.ask("Create project in the server?", default=True)
            if create:
                dc = client.diskclient.DiskClient(url_service, wf_state=state)
                dc.logincli()
                created = init_script.create_on_the_server(root, dc, state)
                if created:
                    init_script.final_words(created.pd.name)
                    init_script.refresh_project(root, created, url_service)
                else:
                    confirm = Confirm.ask(
                        "Do you want to try another name for the project?", default=True
                    )
            else:
                created = True
                init_script.final_words(state.project_name)


@click.command()
@click.option(
    "--url-service",
    "-u",
    default=None,
    help="URL of the NB Workflow Service",
)
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
def info(url_service, from_file):
    """General info and status of the client"""
    # click.echo(f"\nLogin to NB Workflows services {url_service}\n")
    if Path(from_file).is_file():
        c = client.from_file()
    else:
        settings = load_client()
        url_service = settings.WORKFLOW_SERVICE
        c = client.from_env(settings)

    print_json(data=c.info())



from datetime import datetime
from functools import wraps
from inspect import isawaitable
from typing import Optional, Union

from pydantic.error_wrappers import ValidationError
from sanic import Request
from sqlalchemy import delete
from sqlalchemy import insert as sqlinsert
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import selectinload

from labfunctions import defaults
from labfunctions.hashes import generate_random
from labfunctions.models import ProjectModel, UserModel, assoc_projects_users
from labfunctions.security import AuthSpec, PasswordScript, get_delta
from labfunctions.security.errors import (
    AuthValidationFailed,
    MissingAuthorizationHeader,
    WebAuthFailed,
)
from labfunctions.types.security import JWTResponse, UserLogin
from labfunctions.types.user import UserOrm


def select_user():
    stmt = select(UserModel).options(
        selectinload(UserModel.projects),
    )
    # stmt = select(UserModel)
    return stmt


def join_with_project(username):
    stmt = select_user().join(ProjectModel.users).where(UserModel.username == username)
    return stmt


def join_by_project():
    stmt = select_user().join(ProjectModel.users)
    return stmt


def model2orm(user: UserModel) -> UserOrm:
    projects = []
    if user.projects:
        projects = [p.name for p in user.projects]

    return UserOrm(
        username=user.username,
        id=user.id,
        email=user.email,
        password=user.password,
        scopes=user.scopes,
        is_superuser=user.is_superuser,
        is_agent=user.is_agent,
        projects=projects,
    )


def _insert(user: UserOrm):
    t = UserModel.__table__
    stmt = (
        sqlinsert(t).values(**user.dict(exclude={"projects", "id"})).returning(t.c.id)
    )
    return stmt


def _insert_project_user(project_id, user_id):
    stmt = sqlinsert(assoc_projects_users).values(project_id, user_id)


def encrypt_password(password, salt: Union[bytes, str]):

    pm = PasswordScript(salt)
    encrypted = pm.encrypt(password)
    return encrypted


def create(
    session, u: UserOrm, password: Optional[str] = None, salt: Optional[str] = None
) -> Union[int, None]:
    """
    it will create a user in  database
    if created will return the id, else None
    """
    if u.is_superuser and "admin:read:write" not in u.scopes:
        u.scopes = f"{u.scopes},admin:r:w"

    if password and salt:
        u.password = encrypt_password(password, salt)

    stmt = _insert(u)
    try:
        # um = UserModel(**u.dict())
        r = session.execute(stmt)
        id = r.scalar()
        return id
    except IntegrityError:
        pass

    return None


async def get_user_async(
    session, username: str, projectid: Optional[str] = None
) -> Union[UserModel, None]:
    stmt = select_user().where(UserModel.username == username).limit(1)
    if projectid:
        stmt = join_with_project(username).limit(1)

    rsp = await session.execute(stmt)
    return rsp.scalar_one_or_none()


async def get_userid_async(session, user_id: str) -> Union[UserModel, None]:
    stmt = select_user().where(UserModel.id == user_id).limit(1)

    rsp = await session.execute(stmt)
    return rsp.scalar_one_or_none()


def get_user(session, username: str) -> Union[UserModel, None]:
    stmt = select(UserModel).where(UserModel.username == username).limit(1)
    rsp = session.execute(stmt)

    return rsp.scalar_one_or_none()


def create_or_update(session, user: UserOrm) -> UserModel:
    m = UserModel(**user.dict())

    session.merge(m)
    return m


def disable_user(session, username: str) -> Union[UserModel, None]:
    user = get_user(session, username)
    if user:
        user.is_active = False
        user.updated_at = datetime.utcnow()
        session.add(user)

        return user
    return None


async def delete_user_async(session, username: str):
    stmt = delete(UserModel).where(UserModel.username == username)
    await session.execute(stmt)


def delete_user(session, username: str):
    stmt = delete(UserModel).where(UserModel.username == username)
    session.execute(stmt)


async def disable_user_async(session, username: str) -> Union[UserModel, None]:
    user = await get_user_async(session, username)
    if user:
        user.is_active = False
        user.updated_at = datetime.utcnow()
        session.add(user)

        return user
    return None


async def change_pass_async(session, username: str, new_password: str, salt):
    u = await get_user_async(session, username)
    if u:
        pass_ = encrypt_password(new_password, salt=salt)
        u.password == pass_
        u.updated_at = datetime.utcnow()
        session.add(u)


def change_pass(session, user: str, new_password: str, salt) -> bool:
    obj = get_user(session, user)
    if obj:
        pass_ = encrypt_password(new_password, salt=salt)
        obj.password = pass_
        obj.updated_at = datetime.utcnow()
        session.add(obj)
        return True
    return False


def verify_password(
    user: UserModel, pass_unencrypted: str, salt: Union[bytes, str]
) -> bool:
    pm = PasswordScript(salt)
    verified = pm.verify(pass_unencrypted, user.password)
    if verified and user.is_active:
        return True
    return False


async def authenticate(request: Request, *args, **kwargs):
    try:
        creds = UserLogin(**request.json)
    except ValidationError as e:
        raise AuthValidationFailed()

    session = request.ctx.session
    async with session.begin():
        user = await get_user_async(session, creds.username)
        if user is None:
            raise AuthValidationFailed()

        is_valid = verify_password(
            user, creds.password, salt=request.app.config.AUTH_SALT
        )
        if not is_valid:
            raise AuthValidationFailed()

        return user


async def get_jwt_token(auth: AuthSpec, user: UserModel, exp=30) -> JWTResponse:
    access_token = auth.encode(
        {"usr": user.username, "scopes": user.scopes.split(",")}, exp=get_delta(exp)
    )
    username = str(user.username)
    rftkn = await auth.store_refresh_token(username)
    return JWTResponse(access_token=access_token, refresh_token=rftkn)


def inject_user(func):
    """Inject a user"""

    def decorator(f):
        @wraps(f)
        async def decorated_function(request, *args, **kwargs):
            token = request.ctx.token_data
            session = request.ctx.session
            user = get_user_async(session, token["usr"])
            if isawaitable(user):
                user = await user
            if not user:
                raise WebAuthFailed("Authentication failed")
            user_orm = model2orm(user)
            response = f(request, user=user_orm, *args, **kwargs)
            if isawaitable(response):
                response = await response

            return response

        return decorated_function

    return decorator(func)



import os
import sys
from datetime import datetime
from pathlib import Path

import click
import httpx
from rich import print_json

# from labfunctions.io.fileserver import FileFileserver
from rich.console import Console
from rich.table import Table

from labfunctions import client, defaults
from labfunctions.client import init_script
from labfunctions.conf import load_client
from labfunctions.utils import format_seconds, mkdir_p

console = Console()


def get_notebook(nbclient, row):
    output_result = False
    if row.status == 0:
        uri = f"{row.result.output_dir}/{row.result.output_name}"
        if row.result.output_dir:
            mkdir_p(row.result.output_dir)
            output_result = True
    else:
        uri = f"{row.result.error_dir}/{row.result.output_name}"
        if row.result.output_dir:
            mkdir_p(row.result.error_dir)
            output_result = True
    if not Path(uri).exists() and output_result:
        try:
            with open(uri, "wb") as f:
                for chunk in nbclient.history_get_output(uri):
                    f.write(chunk)
        except Exception as e:
            console.print(f"[bold red]Error getting result from {uri}[/]")
            console.print(f"[bold red]{e}[/]")
    return output_result, uri


@click.group(name="status")
def statuscli():
    """
    History of tasks executed
    """


@statuscli.command(name="history")
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--url-service",
    "-u",
    default=load_client().WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option("--wfid", "-w", default=None, help="Execution history of workflow id")
@click.option("--last", "-l", default=1, help="The last executions")
def historycli(from_file, url_service, last, wfid):
    """History of tasks executions"""
    c = client.from_file(from_file, url_service=url_service)

    rsp = c.history_get_last(wfid, last)
    table = Table(title="History")
    # table.add_column("alias", style="cyan", no_wrap=True, justify="center")
    table.add_column("wfid", style="cyan", justify="center")
    table.add_column("execid", style="cyan", justify="center")
    table.add_column("status", style="cyan", justify="center")
    table.add_column("dir_output", style="cyan", justify="center")
    table.add_column("runned", style="cyan", justify="center")

    # print("wfid | execid | status")
    for r in rsp:
        status = "[bold green]OK[/]" if r.status == 0 else "[bold red]FAIL[/]"
        pid = r.result.projectid
        dt = datetime.fromisoformat(r.created_at)
        now = datetime.utcnow()
        diff = (now - dt).total_seconds()
        run = f"[blue]{format_seconds(diff)}[/]"
        c.history_nb_output
        out, uri = get_notebook(c, r)
        if not out:
            uri = "[red bold]Output not generated[/]"

        # print(f"{r.wfid} | {r.execid} | {status} | {uri}")
        table.add_row(r.wfid, r.execid, status, uri, run)

    console.print(table)


@statuscli.command(name="log")
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--url-service",
    "-u",
    default=load_client().WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option("--execid", "-e", default=None, help="Execution log of a task")
@click.option("--nice", "-n", is_flag=True, default=True, help="Print nice output")
def logcli(url_service, from_file, execid, nice):
    """log detail of a execution"""
    c = client.from_file(from_file, url_service=url_service)
    rsp = c.history_detail(execid)
    if not rsp:
        console.print(f"[red bold](x) Execution id {execid} not found[/]")
        sys.exit(-1)
    print_json(data=rsp.dict())
    if nice:
        console.print(f"[red]{rsp.result.error_msg}[/]")



from dataclasses import dataclass
from typing import Dict, List, Optional

from pydantic import BaseModel

from .core import NBTask, WorkflowDataWeb
from .projects import ProjectData


class WorkflowsFile(BaseModel):
    project: ProjectData
    version: str = "0.2"
    workflows: Optional[Dict[str, WorkflowDataWeb]] = {}


@dataclass
class WFCreateRsp:
    status_code: int
    msg: Optional[str] = None
    wfid: Optional[str] = None


@dataclass
class ScheduleExecRsp:
    status_code: int
    msg: Optional[str] = None
    execid: Optional[str] = None


@dataclass
class ScheduleListRsp:
    nb_name: str
    wfid: str
    enabled: bool
    description: Optional[str] = None


@dataclass
class WorkflowRsp:
    enabled: bool
    task: NBTask



from typing import Any, Dict, List, Optional, Union

from sqlalchemy import select
from sqlalchemy.orm import selectinload

from labfunctions.models import HistoryModel
from labfunctions.types import (
    ExecutionResult,
    HistoryLastResponse,
    HistoryResult,
    NBTask,
)


def select_history():

    stmt = select(HistoryModel).options(selectinload(HistoryModel.project))
    return stmt


async def get_last(
    session, projectid: str, wfid: Optional[str] = None, limit=1
) -> Union[HistoryLastResponse, None]:
    if wfid:
        stmt = (
            select(HistoryModel)
            .where(HistoryModel.wfid == wfid)
            .where(HistoryModel.project_id == projectid)
            .order_by(HistoryModel.created_at.desc())
            .limit(limit)
        )
    else:
        stmt = (
            select(HistoryModel)
            .where(HistoryModel.project_id == projectid)
            .order_by(HistoryModel.created_at.desc())
            .limit(limit)
        )

    r = await session.execute(stmt)
    results = r.scalars()
    if not results:
        return None

    rsp = []
    for r in results:
        rsp.append(
            HistoryResult(
                wfid=r.wfid,
                execid=r.execid,
                status=r.status,
                result=r.result,
                created_at=r.created_at.isoformat(),
            )
        )
    return HistoryLastResponse(rows=rsp)


async def get_one(session, execid: str) -> Union[HistoryResult, None]:
    stmt = select(HistoryModel).where(HistoryModel.execid == execid).limit(1)
    r = await session.execute(stmt)
    model: Union[HistoryModel, None] = r.scalar_one_or_none()
    hr = None
    if model:
        hr = HistoryResult(
            wfid=model.wfid,
            execid=model.execid,
            status=model.status,
            result=model.result,
            created_at=model.created_at.isoformat(),
        )
    return hr


async def create(session, execution_result: ExecutionResult) -> HistoryModel:
    result_data = execution_result.dict()

    status = 0
    if execution_result.error:
        status = -1

    row = HistoryModel(
        wfid=execution_result.wfid,
        execid=execution_result.execid,
        project_id=execution_result.projectid,
        elapsed_secs=execution_result.elapsed_secs,
        nb_name=execution_result.name,
        result=result_data,
        status=status,
    )
    session.add(row)
    return row



from .client import LoginError, ProjectCreateError
from .generics import (
    AuthValidationFailed,
    CommandExecutionException,
    PrivateKeyNotFound,
    ProjectNotFound,
    WorkflowDisabled,
    WorkflowNotFound,
)



import getpass
import json
import logging
from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import httpx
from rich.console import Console
from rich.panel import Panel

from labfunctions import defaults, errors, secrets
from labfunctions.conf.jtemplates import render_to_file
from labfunctions.types import (
    ExecutionResult,
    HistoryRequest,
    HistoryResult,
    NBTask,
    ProjectData,
    ProjectReq,
    ScheduleData,
    WorkflowData,
    WorkflowDataWeb,
    WorkflowsList,
)
from labfunctions.utils import parse_var_line

from .base import BaseClient
from .history_client import HistoryClient
from .projects_client import ProjectsClient
from .utils import (
    get_credentials_disk,
    get_private_key,
    store_credentials_disk,
    store_private_key,
)
from .workflows_client import WorkflowsClient


def get_params_from_nb(nb_dict):
    params = {}
    for cell in nb_dict["cells"]:
        if cell["cell_type"] == "code":
            if cell["metadata"].get("tags"):
                _params = cell["source"]
                for p in _params:
                    try:
                        k, v = parse_var_line(p)
                        params.update({k: v})
                    except IndexError:
                        pass
    return params


def open_notebook(fp) -> Dict[str, Any]:
    with open(fp, "r", encoding="utf-8") as f:
        data = json.loads(f.read())
    return data


console = Console()


class DiskClient(WorkflowsClient, ProjectsClient, HistoryClient, BaseClient):
    """Is to be used as cli client because it has side effects on local disk"""

    def __init__(self, *args, **kwargs):
        super(DiskClient, self).__init__(*args, **kwargs)
        self.console = Console()

    def login(self, u: str, p: str):
        super().login(u, p)
        store_credentials_disk(self.creds, self.homedir)

    def logincli(self):
        creds = get_credentials_disk(self.homedir)
        if creds:
            self.creds = creds
            rsp = self.verify()
            if rsp:
                return True
        self.creds = None
        console.print(f"You are connecting to [magenta]{self._addr}[/magenta]")
        u = input("User: ")
        p = getpass.getpass("Password: ")
        self.login(u, p)

    def projects_private_key(self) -> str:
        """Gets private key to be shared to the docker container of a
        workflow task
        """
        try:
            r = self._http.get(f"/projects/{self.projectid}/_private_key")
        except httpx.ConnectError:
            raise errors.PrivateKeyNotFound(self.projectid)

        key = None
        if r.status_code == 200:
            key = r.json().get("private_key")
        if not key:
            raise errors.PrivateKeyNotFound(self.projectid)

        store_private_key(key, self.working_area)
        return key

    def get_private_key(self) -> str:
        """shortcut for getting a private key locally
        TODO: separate command line cli from a general client and an agent client
        a command line cli has filesystem side effects and a agent client not"""
        key = get_private_key(self.working_area)
        if not key:
            return self.projects_private_key()
        return key

    @staticmethod
    def notebook_template(fp):
        render_to_file("test_workflow.ipynb.j2", fp)

    def create_workflow(self, nb_fullpath, alias):
        p = Path(nb_fullpath)
        if p.is_file():
            nb_dict = open_notebook(p)
            params = get_params_from_nb(nb_dict)
            if not params:
                self.logger.warning(f"Params not found for {nb_fullpath}")
        else:
            self.logger.warning(f"Notebook {nb_fullpath} not found, I will create one")
            self.notebook_template(nb_fullpath)

        nb_name = nb_fullpath.rsplit("/", maxsplit=1)[1].split(".")[0]

        task = NBTask(nb_name=nb_name, params=params)
        wd = WorkflowDataWeb(
            alias=alias,
            nbtask=task,
            enabled=False,
        )

        self.state.add_workflow(wd)
        self.write()

    # def info(self):
    #    self.console.print(f"[bold]Project ID: [magenta]{self.projectid}[/][/]")
    #    self.console.print(f"[bold]Project Name: [blue]{self.project_name}[/][/]")
    # self.console.print(table)



# Shortcuts for client or agents only
from labfunctions import secrets
from labfunctions.client.shortcuts import from_env, from_file
from labfunctions.conf.utils import load_client as conf_loader

settings = conf_loader()
secrets = secrets.load(settings.BASE_PATH)
client = from_env(settings)
